﻿//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  Cy4NET ASCII PDU (Server Side Edition)                                              //
//  _____________________________________________________                               //
//                                                                                      //
//  Copyright (c) 2010 Reimar Grasbon													//
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  This software is subjected to the FreeBSD License. Redistribution and use in source //
//  and binary forms, with or without modification, are permitted provided that the     //
//  following conditions are met:                                                       //
//                                                                                      //
//  1. Redistributions of source code must retain the above copyright notice, this list //
//     of conditions and the following disclaimer.                                      //
//                                                                                      //
//  2. Redistributions in binary form must reproduce the above copyright notice, this   //
//     list of conditions and the following disclaimer in the documentation and/or      //
//     other materials provided with the distribution.                                  //
//                                                                                      //
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY //
//  EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED           //
//  WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  //
//  IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,    //
//  INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT  //
//  NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR  //
//  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,   //
//  WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)  //
//  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE          //
//  POSSIBILITY OF SUCH DAMAGE.                                                         //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//   Cy4 ASCII PDU Synopsis:                                                            //
//   -----------------------                                                            //
//                                      XX XX XX XXXXXX XX XX XX ... XX XX              //
//                                      |  |  |    |    |  |  |      |  |               //
//                                      |  |  |    |    |  |  |      |  |               //
//           DTF:   "C0"   ..   "C5" ---+  |  |    |    |  |  |      |  |               //
//           NET:   "00"   ..   "FF" ------+  |    |    |  |  |      |  |               //
//           DEV:   "00"   ..   "FF" ---------+    |    |  |  |      |  |               //
//           LOC: "000000" .. "FFFFFF" ------------+    |  |  |      |  |               //
//           CNT:   "00"   ..   "40" -------------------+  |  |      |  |               //
//           DAT 1: "00"   ..   "FF" ----------------------+  |      |  |               //
//           DAT 2: "00"   ..   "FF" -------------------------+      |  |               //
//            :      ::    ..    ::                                  |  |               //
//           DAT n: "00"   ..   "FF" --------------------------------+  |               //
//           CRC:   "00"   ..   "FF" -----------------------------------+               //
//                                                                                      //
//                                                                                      //
//   · DTF: Datagramm Type Flags Byte:                                                  //
//                                                                                      //
//           7 6 5 4 3 2 1 0                                                            //
//          +-+-+-+-+-+-+-+-+                                                           //
//          |1|1|0|0|0|F|D|P|                                                           //
//          +-+-+-+-+-+-+-+-+                                                           //
//                     | | |                                                            //
//                     | | +--- P: Payload flag:   1 = Has payload  , 0 = No payload    //
//                     | +----- D: Direction flag: 1 = Request      , 0 = Response      //
//                     +------- F: Failure flag:   1 = Access failed, 0 = Correct       //
//                                                                                      //
//                            DTF-Values    |     Type                                  //
//                       -------------------+----------------                           //
//                          11000000b (C0h) | Write Response                            //
//                          11000001b (C1h) | Read  Response                            //
//                          11000010b (C2h) | Read  Request                             //
//                          11000011b (C3h) | Write Request                             //
//                          11000100b (C4h) | Write Failure                             //
//                          11000101b (C5h) | Read  Failure                             //
//                                                                                      //
//   · NET: Net address of the destination node.                                        //
//                                                                                      //
//   · DEV: Device address of the destination node.                                     //
//                                                                                      //
//   · LOC: Unsigned 24 bit area number of the requestes service.                       //
//                                                                                      //
//   · CNT: Dependend on the type of datagramm: Number of provided (write request),     //
//          written (write response), requested (read request) or read (read response)  //
//          data bytes. Maximum allowed value here is 64 (40h).                         //
//                                                                                      //
//   · DAT: Payload. Only datagrams of type 'write request' and 'read response' have    //
//          payload bytes in section DAT. In Write responses, read requests and failed  //
//          datagrams have no DAT bytes, section DAT is skipped.                        //
//                                                                                      //
//   · CRC  8-Bit Dow-CRC across all PDU bytes, except CRC.                             //
//                                                                                      //
//                                                                                      //
//                                                                                      //
//  Example 1: Query information from device @ 255.3                                    //
//                                                                                      //
//      Read request:     "C2FF03FFFFF00717"                                            //
//      Read response:    "C1FF03FFFFF007A0012008B00023EA"                              //
//                                                                                      //
//                                                                                      //
//  Example 2: Restart device @ 1.6                                                     //
//                                                                                      //
//      Write request:    "C30106FFFFF8015EC6"                                          //
//      Write response:   "C00106FFFFF80175"                                            //
//                                                                                      //
//                                                                                      //
//  Example 3: Query information from nonexisting device @ 20.3                         //
//                                                                                      //
//      Read request:     "C21403FFFFF007C6"                                            //
//      Read failure:     "C51403FFFFF00775"                                            //
//                                                                                      //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////| RG |//

#include "Conf.h"
#include "Cy4Net.h"
#include "AsciiPdu.h"

//////////////////////////////////////////////////////////////////////////////////////////
//  Lokale Funktionen  ///////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
static void
UpdateCRC(UInt08 data, UInt08* crc)
{
    *crc ^= data;
    for (UInt08 i=0; i<8; i++) *crc = (*crc >> 1) ^ ((*crc & 1) ? 0x8C : 0 );
}

//////////////////////////////////////////////////////////////////////////////////////////
static void
PrintUInt08Hex(char** txt, UInt08 val)
{
    for (UInt08 i=0; i<2; i++)
    {   UInt08 n = (unsigned char)((val>>4) & 0xF);
        if (n < 10) **txt = (char)('0'    + n);
        else        **txt = (char)('A'-10 + n);
        (*txt)++;
        val<<=4;
    }
    **txt = '\0';
}

//////////////////////////////////////////////////////////////////////////////////////////
static bool
ParseUInt08Hex(const char** text, UInt08* value)
{
    const char* tmpText = *text;
    *value = 0;
    for (UInt08 i=0; i<2; i++)
    {   UInt08 c = (UInt08)(*tmpText);
        if (c >= 'a') c -= 'a' - 'A';  // Großbuchstaben
        UInt08 n;
        if      (c >= '0' && c <= '9') n =  0 + c - '0';
        else if (c >= 'A' && c <= 'F') n = 10 + c - 'A';
        else                           return false;
        *value = *value*16 + n;
        tmpText++;
    }
    *text = tmpText;
    return true;
}

//////////////////////////////////////////////////////////////////////////////////////////
//  Funktionen  //////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
bool
ASCIIPDU_ParseRequest(const char* line, Cy4Message* message)
{
    CY4MESSAGE_Invalidate(message);
    if (line == NULL || line[0] == '\0') return false;
    if (line[0] == '*' && line[1] == '\0')
    {   // Auf Ping-Anfrage "*" Geräte-Typenschild zurückliefern
        CY4MESSAGE_SetupRead(message, CY4NET_Address(), 0xFFFFF0, 7);
        return true;
    }
    UInt08 crc;
    crc = 0xFF;
    UInt08 buffer[7];
    UInt08 index;
    for (index=0; index<7; index++)
    {   if (!ParseUInt08Hex(&line, &buffer[index])) return false;
        UpdateCRC(buffer[index], &crc);
    }
    if (buffer[0] < 0xC2 || buffer[0] > 0xC3) return false;    // Nur Anfragen erlaubt
    bool payload = buffer[0] & 0x01;
    Cy4MessageAccess  access;
    if (payload) access = CY4MESSAGE_WRITE;
    else         access = CY4MESSAGE_READ;
    Cy4Address address = CY4ADDRESS(buffer[1], buffer[2]);
    UInt32 location = ((UInt32)buffer[3]<<16) + ((UInt16)buffer[4]<<8) + buffer[5];
    UInt08 count = buffer[6];
    if (count > CY4MESSAGE_CAPACITY) return false;
    UInt08 data[CY4MESSAGE_CAPACITY];
    if (payload)
    {   for (index=0; index<count; index++)
        {   if (!ParseUInt08Hex(&line, &data[index])) return false;
            UpdateCRC(data[index], &crc);
        }
        index += 7;
    }
	UInt08 checkSum;
    if (!ParseUInt08Hex(&line, &checkSum)) return false;
    if (*line != '\0' || crc != checkSum) return false;		
    if (access == CY4MESSAGE_READ)
    {   CY4MESSAGE_SetupRead(message, address, location, count);
    }
    else
    {   CY4MESSAGE_SetupWrite(message, address, location, data, count);
    }
    return true;
}

//////////////////////////////////////////////////////////////////////////////////////////
void
ASCIIPDU_PrintResponse(const Cy4Message* message, char line[144+1])
{
    char* text = &line[0];
    bool payload   = CY4MESSAGE_Access(message) == CY4MESSAGE_READ;
    bool direction = false;  // Response
    bool failure   = !CY4MESSAGE_IsValid(message);
    UInt08  buffer[7];
    buffer[0] = 0xC0 + (failure<<2) + (direction<<1) + (payload<<0);		
	Cy4Address address = CY4MESSAGE_Address(message);
    buffer[1] = CY4ADDRESS_Net(&address);
    buffer[2] = CY4ADDRESS_Dev(&address);
	UInt32 location = CY4MESSAGE_Location(message);
    buffer[3] = (UInt08)(location >> 16);
    buffer[4] = (UInt08)(location >>  8);
    buffer[5] = (UInt08)(location >>  0);
    buffer[6] = CY4MESSAGE_Count(message);
    UInt08 crc = 0xFF;
    for (UInt08 i=0; i<7; i++)
    {   PrintUInt08Hex(&text, buffer[i]);
        UpdateCRC(buffer[i], &crc);
    }
    if (!failure && payload)
    {   for (UInt08 i=0; i<buffer[6]; i++)
        {   PrintUInt08Hex(&text, CY4MESSAGE_ConstData(message)[i]);
            UpdateCRC(CY4MESSAGE_ConstData(message)[i], &crc);
        }
    }
    PrintUInt08Hex(&text, crc);
}

//////////////////////////////////////////////////////////////////////////////////////////
