﻿//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  Boardmodul: Bereitstellung gerätespezifischer Funktionen und Mechanismen            //
//  ________________________________________________________________________            //
//                                                                                      //
//  Copyright (c) 2019 Reimar Grasbon												    //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  This software is subjected to the FreeBSD License. Redistribution and use in source //
//  and binary forms, with or without modification, are permitted provided that the     //
//  following conditions are met:                                                       //
//                                                                                      //
//  1. Redistributions of source code must retain the above copyright notice, this list //
//     of conditions and the following disclaimer.                                      //
//                                                                                      //
//  2. Redistributions in binary form must reproduce the above copyright notice, this   //
//     list of conditions and the following disclaimer in the documentation and/or      //
//     other materials provided with the distribution.                                  //
//                                                                                      //
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY //
//  EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED           //
//  WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  //
//  IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,    //
//  INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT  //
//  NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR  //
//  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,   //
//  WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)  //
//  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE          //
//  POSSIBILITY OF SUCH DAMAGE.                                                         //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////

#include <avr/interrupt.h>
#include <avr/wdt.h>
#include <avr/pgmspace.h>
#include <avr/eeprom.h>
#include <util/twi.h>
#include "Conf.h"
#include "Board.h"
#include "Timer.h"

//////////////////////////////////////////////////////////////////////////////////////////
//  Definitionen  ////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
#define SYS_LED     A,7             // -+
#define SER_ACT_LED C,7             //  |
#define BUS_ACT_LED C,6             //  |
#define GP_LED3     D,5             //  | Aufgabenzuordnung der Port-Pins
#define GP_LED2     D,6             //  |
#define TA1         B,1             //  |
#define TA2         B,2             // -+

//////////////////////////////////////////////////////////////////////////////////////////
//  Typen  ///////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
typedef enum                        // Arbeitsanweisungen für das Board-Modul:
{   DO_STARTUP,                     // · Starte die Anwendung neu
    DO_ENTER_PGMLDR,                // · Starte den Programmlader
    DO_PROCEED,                     // · Führe die Arbeit fort
    DO_SHUTDOWN                     // · Schalte das Board ab
} DoNext;

typedef struct                      // Bus-Übertragungsstatus:
{   bool   pending : 1;             // · Übertragung im Gange
    bool   echo    : 1;             // · Echo erhalten
    bool   error   : 1;             // · Übertragung Fehlgeschlagen (Kollision)
    UInt08 byteGap : 5;             // · Vergangene Zeit seit dem letzten Byteempfang [ms]
} BusStatus;

typedef struct                      // Element der Bus-Empfangsschlange:
{   bool   timeout : 1;             // · Zeitüberschreitung bei 'byteGap' aufgetreten
    UInt08 padding : 7;             // · Füllbits (ohne Bedeutung)
    UInt08 data    : 8;             // · Datenbyte
} BusRxItem;

//////////////////////////////////////////////////////////////////////////////////////////
//  Daten  ///////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
static volatile UInt16       msecTicks                             = 0;
static volatile UInt16       secTicks                              = 0;
static volatile UInt08       busTxQueue[CFG_BUS_TX_QUEUE_CAPACITY] = { 0 };
static volatile UInt08       busTxQueueHead                        = 0;
static volatile UInt08       busTxQueueLength                      = 0;
static volatile UInt08       busLoopbackByte                       = 0;
static volatile BusStatus    busStatus                             = { false, false, false, 0};
static volatile BusRxItem    busRxQueue[CFG_BUS_RX_QUEUE_CAPACITY] = { { false } };
static volatile UInt08       busRxQueueHead                        = 0;
static volatile UInt08       busRxQueueTail                        = 0;
static volatile UInt08       serTxQueue[CFG_SER_TX_QUEUE_CAPACITY] = { 0 };
static volatile UInt08       serTxQueueHead                        = 0;
static volatile UInt08       serTxQueueLength                      = 0;
static volatile UInt08       serRxQueue[CFG_SER_RX_QUEUE_CAPACITY] = { 0 };
static volatile UInt08       serRxQueueHead                        = 0;
static volatile UInt08       serRxQueueLength                      = 0;
static          DoNext       doNext           				       = DO_STARTUP;
static          DateTime     sysDateTime                           = DATETIME_INVALID;
static          UInt16       sysTimeTicks                          = 0;
static          UInt08       busActLedOn                           = 0;
static          UInt08       serActLedOn                           = 0;
static          bool         keyPressedEvent[2]                    = { false, false };
static          bool         keyPressedState[2]                    = { false, false };
static          SysLedSignal sysLed                                = SYS_LED_OFF;
static          bool         toneMute                              = true;
static			UInt08       toneSustain                           = 0;

//////////////////////////////////////////////////////////////////////////////////////////
//  Lokale Funktionen  ///////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////////////////////////
//  Reset(): Führt einen physikalischen Reset herbei indem der Watchdog ausgelöst wird. //
//  -------  Reset() kehrt nie zum Aufrufer zurück.                                     //
//                                                                                      //
//  Eingabe: -                                                                          //
//  Ausgabe: -                                                                          //
//////////////////////////////////////////////////////////////////////////////////////////
__attribute__((noreturn)) static void
Reset(void)
{
    DisableInterrupts();    // Interrupts sperren
    wdt_enable(WDTO_15MS);  // Watchdog-Intervall auf kürzeste Zeit setzen
    while (true);           // Auf "Zubiss" des Watchdogs warten
}

//////////////////////////////////////////////////////////////////////////////////////////
static void
EnterPgmLdr(void)
{
    if (pgm_read_word(FLASHEND - CFG_PGMLDR_SIZE*1024 + 1) == 0x940C)
    {   GOTO(FLASHEND - CFG_PGMLDR_SIZE*1024 + 1+4);
    }
}

//////////////////////////////////////////////////////////////////////////////////////////
static UInt16
ReadADC(UInt08 channel) // Kanalnummer: 0..7
{
//  ADMUX =             channel;  // Eingangsspannung bezogen auf VRef
    ADMUX = BV(REFS0) | channel;  // Eingangsspannung bezogen auf AVcc
    ADCSRA |=  BV(ADSC);
    loop_until_bit_is_clear(ADCSRA, ADSC);
    UInt08 lo  = ADCL;
    UInt08 hi  = ADCH;
    return ((UInt16)(hi & 0x03) << 8) | lo;   // Ausgabe: 0..1023
}

//////////////////////////////////////////////////////////////////////////////////////////
static void
ProceedKeying(void)
{
    // 1ms Zeitbarriere
	UInt16 ticks = BOARD_MSecTicks();
	{	static UInt08 ms = 0;
		if (ms == (UInt08)ticks) return;
		ms = (UInt08)ticks;
    }
    // Tastenkontrolle TA1 und TA2
    for (UInt08 i=0; i<2; i++)
    {   static UInt08 latency[2] = { 0, 0 };
        bool isPressed;
        switch (i)
        {   case 0:  isPressed = !GetPin(TA1); break;
            case 1:  isPressed = !GetPin(TA2); break;
            default: isPressed = false;
        }
        if (keyPressedState[i])
        {   // Taster geschlossen
            if (!isPressed && latency[i] > 0) latency[i]--;
            else                              latency[i] = CFG_PB_UP_RESPONSIVITY;
        }
        else
        {   // Taster geöffnet
            if (isPressed && latency[i] > 0) latency[i]--;
            else                             latency[i] = CFG_PB_DOWN_RESPONSIVITY;
        }
        if (latency[i] == 0)
        {   keyPressedState[i] = isPressed;
            keyPressedEvent[i] = isPressed;
        }
    }
}

//////////////////////////////////////////////////////////////////////////////////////////
static void
ProceedLedSignaling(void)
{
    // 1ms Zeitbarriere
	UInt16 ticks = BOARD_MSecTicks();
	{	static UInt08 ms = 0;
		if (ms == (UInt08)ticks) return;
		ms = (UInt08)ticks;
    }
    // Darstellung der System-LED
    switch (sysLed)
    {   case SYS_LED_ON:
        {   SetPin(SYS_LED);
            break;
        }
        case SYS_LED_OFF:
        {   ClrPin(SYS_LED);
            break;
        }
        case SYS_LED_HEARTBEAT:
        {   ChgPin(SYS_LED, (bool)((ticks&0x0380)==0x0080 || (ticks&0x0380)==0x0180));
            break;
        }
        default:
            ;
    }
    // Bus-Aktivität
    if (busActLedOn > 0)
    {   busActLedOn--;
        ChgPin(BUS_ACT_LED, busActLedOn > 0);
    }
    // Serielle-Aktivität
    if (serActLedOn > 0)
    {   serActLedOn--;
        ChgPin(SER_ACT_LED, serActLedOn > 0);
    }
}

//////////////////////////////////////////////////////////////////////////////////////////
static void
ProceedSysDateTiming(void)
{
    if (sysDateTime.time.second < 60)
    {   UInt16 ticks = BOARD_MSecTicks();
        if (sysTimeTicks+999 < ticks)
        {   sysTimeTicks += 1000;
            DATETIME_ShiftSecond(&sysDateTime, 1);
        }
    }
}

//////////////////////////////////////////////////////////////////////////////////////////
static void
ProceedTonePlaying(void)
{
    static Timer timer = TIMER_MILLISECONDS(10);

    if (toneSustain > 0)
    {   if (TIMER_RestartIfExpired(&timer))
        {   toneSustain--;
            if (toneSustain == 0)
            {   TIMER_Stop(&timer);
                BOARD_PlayTone(NOTE_XX, 0);
            }
		}
    }
}

//////////////////////////////////////////////////////////////////////////////////////////
//  BOARD_PreStart(): Diese Funktion wird einmal nach dem Systemstart aufgerufen. Der   //
//  ----------------  Zeitpunkt ihres Aufrufs macht sie zur einer Sonderfunktion, denn  //
//  ihr Aufruf findet in der Start-Sektion "init3", noch lange vor Aufruf von main(),   //
//  statt. Das ist der richtige Zeitpunkt, um z.B. die Ursache des Resets zu ermitteln  //
//  oder statische Variablen auszulesen, deren Werte vor dem Reset platziert wurden     //
//  und den Neustart überlebt haben. Denn zu diesem Zeitpunkt sind die statischen       //
//  Variablen noch nicht initialisiert.                                                 //
//                                                                                      //
//  Eingabe: -                                                                          //
//  Ausgabe: -                                                                          //
//////////////////////////////////////////////////////////////////////////////////////////
__attribute__((naked, section(".init3"))) void
BOARD_PreStart(void)
{
    MCUSR = 0;
    wdt_disable();
}

//////////////////////////////////////////////////////////////////////////////////////////
//  Interruptbehandler  //////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
ISR(TIMER1_COMPA_vect)
{
    // Ticker inkrementieren
    {   msecTicks++;
        static UInt16 counter = 0;
        counter++;
        if (counter == 1000)
        {   counter = 0;
            secTicks++;
        }
    }
    // Bus Byte-Zeitlücke
    if (busStatus.byteGap <= CFG_BUS_INTERBYTE_TIMEOUT)
    {   busStatus.byteGap++;
        if (busStatus.byteGap > CFG_BUS_INTERBYTE_TIMEOUT)
        {   busRxQueue[busRxQueueTail].timeout = true;
        }
    }
}

//////////////////////////////////////////////////////////////////////////////////////////
ISR(USART0_TX_vect)
{                                   // IRQ Abfolge sowie die Signale 'echo' und 'pending':
    if (!busStatus.echo)            //
    {   busStatus.error = true;     // Start Tx:    |1|
        busTxQueueLength = 0;       //               :
    }                               // Tx-IRQ:  ···--:-----|2|----|3|-···-|n|----| |-···
    busStatus.echo = false;         // Rx-IRQ:  ···--:-|1|--:-|2|--:--···-- -|n|--:--···
    UInt08 head = busTxQueueHead;   //               :  :___:  :___:          :___:
    if (busTxQueueLength == 0)      // echo:    ...__:__|   |__|   |____...___|   |__...
    {   busStatus.pending = false;  //               :____________________________:
    }                               // pending: ...__|                            |__... 
    else                                
    {   busLoopbackByte = busTxQueue[head];
        UDR0            = busTxQueue[head];
        head++;
        if (head == CARDINALITY(busTxQueue)) head = 0;
        busTxQueueLength--;
    }
    busTxQueueHead = head;
}

//////////////////////////////////////////////////////////////////////////////////////////
ISR(USART0_RX_vect)
{
    UInt08 byte = UDR0;
    busStatus.byteGap = 0;   // Byte empfangen, Zeitzähler zurücksetzen
    if (busStatus.pending && byte == busLoopbackByte) busStatus.echo = true; // Echo
    // Byte in Empfangsschlange stellen
    UInt08 tail = busRxQueueTail;
    busRxQueue[tail].data = byte;
    tail++;
    if (tail == CARDINALITY(busRxQueue)) tail = 0;
    if (tail == busRxQueueHead) busRxQueue[busRxQueueTail].timeout = true; // Überlauf
    else                        busRxQueueTail = tail;
}

//////////////////////////////////////////////////////////////////////////////////////////
ISR(USART1_UDRE_vect)
{
    UDR1 = serTxQueue[serTxQueueHead];
    serTxQueueHead++;
    if (serTxQueueHead == CARDINALITY(serTxQueue)) serTxQueueHead = 0;
    serTxQueueLength--;
    if (serTxQueueLength == 0) UCSR1B &= ~BV(UDRIE1);
}

//////////////////////////////////////////////////////////////////////////////////////////
ISR(USART1_RX_vect)
{
    UInt08 status = UCSR1A;
    UInt08 byte   = UDR1;
    if (status & (BV(FE1)|BV(DOR1))) return;
    UInt08 tail = serRxQueueHead + serRxQueueLength;
    if (CARDINALITY(serRxQueue) > 127)
    {   if      (tail >= CARDINALITY(serRxQueue)) tail -= CARDINALITY(serRxQueue);
        else if (tail < serRxQueueHead)           tail += 256 - CARDINALITY(serRxQueue);
    }
    else
    {   if (tail >= CARDINALITY(serRxQueue)) tail -= CARDINALITY(serRxQueue);
    }
    serRxQueue[tail] = byte;
    serRxQueueLength++;
}

//////////////////////////////////////////////////////////////////////////////////////////
// Funktionen ////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////////////////////////
//  BOARD_Reset(): Initiiert einen Systemneustart. Die Funktion löst nicht unmittelbar  //
//  -------------  einen Reset aus, sondern führt zu einem Ausstieg aus der Haupt-      //
//  schleife von Main(). Die danach sukzessiv aufgerufenen "Shutdown"-Funktionen aller  //
//  betroffenen Module bewirken den kontrollierten Abschluss aller Aktivitäten, bevor   //
//  die letzte aufgerufene "Shutdown"-Funktion des Board-Moduls dann den physikali-     //
//  schen Reset auslöst.                                                                //
//                                                                                      //
//  Eingabe: -                                                                          //
//  Ausgabe: -                                                                          //
//////////////////////////////////////////////////////////////////////////////////////////
void
BOARD_Reset(void)
{
    doNext = DO_STARTUP;
}

//////////////////////////////////////////////////////////////////////////////////////////
//  BOARD_EnterPgmLdr(): Ähnlich wie BOARD_Reset() initiiert BOARD_EnterPgmLdr() einen  //
//  -------------------  Systemneustart. Im Unterschied zu BOARD_Reset(), wird hier     //
//  jedoch nicht die Anwendung neu gestartet, sondern das System bootet in das Pro-     //
//  grammladermodul.                                                                    //
//                                                                                      //
//  Eingabe: -                                                                          //
//  Ausgabe: -                                                                          //
//////////////////////////////////////////////////////////////////////////////////////////
void
BOARD_EnterPgmLdr(void)
{
    doNext = DO_ENTER_PGMLDR;
}

//////////////////////////////////////////////////////////////////////////////////////////
//  BOARD_MSecTicks(): Die Funktion liefert den Stand eines monotonen Millisekunden-    //
//  -----------------  Zeitzählers. Der Zähler wird interruptgetrieben inkrementiert    //
//  und hat eine Breite von 16-Bit. Kommt es nach 65535 ms zum Überlauf, beginnt der    //
//  Zeitzähler wieder bei 0.                                                            //
//                                                                                      //
//  Eingabe: -                                                                          //
//  Ausgabe: Aktueller Stand des Millisekunden-Zeitzählers.                             //
//////////////////////////////////////////////////////////////////////////////////////////
UInt16
BOARD_MSecTicks(void)
{
    UInt16 ticks = 0;

    Atomic { ticks = msecTicks; }
    return ticks;
}

//////////////////////////////////////////////////////////////////////////////////////////
//  BOARD_SecTicks(): Die Funktion liefert den Stand eines monotonen Sekunden-Zeit-     //
//  ----------------  zählers. Der Zähler wird interruptgetrieben inkrementiert und hat //
//  eine Breite von 16-Bit. Kommt es nach 65535 s (knapp 18¼h) zum Überlauf, beginnt    //
//  der Zeitzähler wieder bei 0.                                                        //
//                                                                                      //
//  Eingabe: -                                                                          //
//  Ausgabe: Aktueller Stand des Sekunden-Zeitzählers.                                  //
//////////////////////////////////////////////////////////////////////////////////////////
UInt16
BOARD_SecTicks(void)
{
    UInt16 ticks = 0;

    Atomic { ticks = secTicks; }
    return ticks;
}

//////////////////////////////////////////////////////////////////////////////////////////
//  BOARD_EEPLoad(): Kopiert 'count' Bytes aus dem EEPROM-Speicher ins RAM.             //
//  ---------------                                                                     //
//                                                                                      //
//  Eingabe: · eepAddress: Adresse im EEPROM-Speicher.                                  //
//           · ramAddress: Adresse im Ram.                                              //
//           · count:      Anzahl zu kopierender Bytes.                                 //
//  Ausgabe: -                                                                          //
//////////////////////////////////////////////////////////////////////////////////////////
void
BOARD_EEPLoad(const void* eepAddress, void* ramAddress, UInt16 count)
{
    eeprom_read_block(ramAddress, (void*)eepAddress, count);
}

//////////////////////////////////////////////////////////////////////////////////////////
//  BOARD_EEPSave(): Kopiert 'count' Bytes aus dem RAM in den EEPROM-Speicher. Da       //
//  ---------------  EEPROM-Speicher lt. Datenblatt nur ca. 100000 Lösch/Schreibzyklen  //
//  pro Zelle verträgt, wird vor für jedes Bytes vor dem Speichervorgang geprüft, ob    //
//  sich der zu schreibende Wert sich vom Inhalt der EEP-Speicherzelle unterscheidet.   //
//  Erst danach wird nötigenfalls ein Lösch- und Schreibvorgang durchgeführt.           //
//  Hinweis: Man beachte, dass das Beschreiben einer EEPROM-Speicherzelle ca. 8 ms in   //
//           Anspruch nimmt und damit verhältnismäßig viel Zeit verschlingt.            //
//                                                                                      //
//  Eingabe: · ramAddress: Adresse im Ram.                                              //
//           · eepAddress: Adresse im EEPROM-Speicher.                                  //
//           · count:      Anzahl zu kopierender Bytes.                                 //
//  Ausgabe: -                                                                          //
//////////////////////////////////////////////////////////////////////////////////////////
void
BOARD_EEPSave(const void* ramAddress, void* eepAddress, UInt16 count)
{
    eeprom_update_block(ramAddress, (void*)eepAddress, count);
}

//////////////////////////////////////////////////////////////////////////////////////////
//  BOARD_BusOnActivity(): Rückruffunktion der Bus-Übertragungseinheit. Die Funktion    //
//  ---------------------  wird aufgerufen, wenn mit dem Gerät über den Bus             //
//  kommuniziert wird.                                                                  //
//                                                                                      //
//  Eingabe: -                                                                          //
//  Ausgabe: -                                                                          //
//////////////////////////////////////////////////////////////////////////////////////////
void
BOARD_BusOnActivity(void)
{
	busActLedOn = CFG_BUS_ACT_LED_FLASH;
}

//////////////////////////////////////////////////////////////////////////////////////////
//  BOARD_BusIsIdle(): Prüft, ob der Bus momentan frei ist. Der Bus gilt als frei, wenn //
//  -----------------  eine Kommunikationspause von mehr als CFG_BUS_INTERBYTE_TIMEOUT  //
//  Millisekunden eintritt.                                                             //
//                                                                                      //
//  Eingabe: -                                                                          //
//  Ausgabe: 'true', wenn der Bus frei ist.                                             //
//////////////////////////////////////////////////////////////////////////////////////////
bool
BOARD_BusIsIdle(void)
{
    bool  idle = false;
    Atomic
    {   idle = !busStatus.pending && busStatus.byteGap > CFG_BUS_INTERBYTE_TIMEOUT;
    }
    return idle;
}

//////////////////////////////////////////////////////////////////////////////////////////
//  BOARD_BusTransmit(): Sendet das Datenbyte in 'byte' auf den Bus.                    //
//  -------------------  Die Sendeeinheit ist durch folgenden Zustandsautomaten         //
//  realisiert, Ausgangszustand ist IDLE:                                               //
//                                                                                      //
//                                                                                      //
//     +----------+ BOARD_BusTransmit()    +------------------+ BOARD_BusTransmit()     //
//     |//////////+----------------------->|//////////////////+--------------------+    //
//     |// IDLE //|              COMPLETED |// TRANSMITTING //|                    |    //
//     |//////////|<-----------------------+//////////////////|<-------------------+    //
//     +----------+                        +--------+---------+                         //
//          /|\                                     | COLLISION                         //
//           |                                      |                                   //
//           |                                     \|/                                  //
//           |                               +--------------+                           //
//           | BOARD_BusTakeFailure()        |//////////////|                           //
//           +-------------------------------+// FAIILURE //|                           //
//                                           |//////////////|                           //
//                                           +--------------+                           //
//                                                                                      //
//  Mit Aufruf dieser Funktion wechselt der Automat IDLE nach TRANSMITTING und beginnt  //
//  mit der Übertragung des angegebenen Datenbytes. Ist bereits eine Übertragung im     //
//  Gange (Zustand TRANSMITTING), wird bei Aufruf der Funktion das zu übertragende      //
//  Datenbyte für die spätere Übertragung in die Sende-Warteschlange gestellt. Nach     //
//  Aussendung aller zu übertragenden Datenbytes (Ereignis COMPLETED) kehrt der Automat //
//  wieder in den Zustand IDLE zurück.                                                  //
//  Das Halbduplexverfahren des seriellen Busses bedingt, dass jedes ausgesendete Byte  //
//  unmittelbar als Echo wieder in den Empfangspuffer zurückwandert. Diese Rücküber-    //
//  tragung wird genutzt, um Kollisionen aufzudecken. Unterscheiden sind ausgesendetes  //
//  Byte und Echo, löst das das Ereignis COLLISION aus und damit den Wechsel in den     //
//  Zustand FAIILURE. Die Übertragung wird in dem Fall nicht mehr fortgesetzt und die   //
//  Sende-Warteschlange geleert. Gegebenenfalls noch ausstehende Bytes werden also      //
//  nicht mehr übertragen. Im Zustand FAIILURE bleiben Aufrufe von BOARD_BusTransmit()  //
//  ohne Wirkung. Um neue Datenbytes senden zu können, muss der Automat erst durch      //
//  Aufruf von BOARD_BusTakeFailure() wieder in den Zustand IDLE versetzt werden.       //
//                                                                                      //
//  Eingabe: · byte: Das zu übertragende Datenbyte.                                     //
//  Ausgabe: -                                                                          //
//////////////////////////////////////////////////////////////////////////////////////////
void
BOARD_BusTransmit(UInt08 byte)
{
    // Ist die Übertragungsschlange voll, warten, bis wieder Platz ist
    bool  queueFull = false;
    do
    {   Atomic
        {   queueFull = busTxQueueLength == CARDINALITY(busTxQueue);
	    }
    } while (queueFull);
    // Byte senden
    Atomic
    {   if (busStatus.error) return; // Zustand FAIILURE: Nichts tun
        if (busStatus.pending)
        {   // Zustand TRANSMITTING: Byte für Übertragung aufreihen
            UInt08 tail = busTxQueueHead + busTxQueueLength;
            if (tail >= CARDINALITY(busTxQueue)) tail -= CARDINALITY(busTxQueue);
            busTxQueue[tail] = byte;
            busTxQueueLength++;
        }
        else
        {   // Zustand IDLE: Byte unmittelbar senden
            busStatus.pending = true;
            busStatus.echo    = false;
            busLoopbackByte   = byte;
            UDR0              = byte;
        }
    }
}

//////////////////////////////////////////////////////////////////////////////////////////
//  BOARD_BusIsTransmitting(): Prüft, ob sich der Automat (siehe BOARD_BusTransmit())   //
//  -------------------------  im Zustand TRANSMITTING befindet. Das ist solange der    //
//  Fall, wie die Sendeeinheit mit der Übertragung von Datenbytes beschäftigt ist.      //
//                                                                                      //
//  Eingabe: -                                                                          //
//  Ausgabe: 'true', wenn der Sendevorgang noch im Gange ist.                           //
//////////////////////////////////////////////////////////////////////////////////////////
bool
BOARD_BusIsTransmitting(void)
{
    bool  pending = false;
    Atomic
    {   pending = busStatus.pending;
    }
    return pending;
}

//////////////////////////////////////////////////////////////////////////////////////////
//  BOARD_BusTakeFailure(): Prüft, ob sich der Automat (siehe BOARD_BusTransmit()) im   //
//  ----------------------  Zustand FAILURE befindet und löst gegebenenfalls einen      //
//  Wechsel in den Zustand IDLE aus. Übertragungsfehler entstehen in der Regel durch    //
//  Kollisionen auf dem Bus.                                                            //
//                                                                                      //
//  Eingabe: -                                                                          //
//  Ausgabe: 'true', wenn ein Übertragungsfehler vorlag.                                //
//////////////////////////////////////////////////////////////////////////////////////////
bool
BOARD_BusTakeFailure(void)
{
    bool  error = false;
    Atomic
    {   error = busStatus.error;
        busStatus.error = false;
    }
    return error;
}

//////////////////////////////////////////////////////////////////////////////////////////
//  BOARD_BusReceive(): Steht bei Aufruf der Funktion ein empfangenes Datenbyte zur     //
//  ------------------  Abholung bereit, wird es aus der Empfangsschlange entfernt und  //
//  in 'byte' übergeben. BOARD_BusReceive() kehrt so lange mit 'true' zurück, wie       //
//  Datenbytes zur Abholung bereitstehen.                                               //
//  Man beachte: Der Rückgabewert von BOARD_BusReceive() sagt nichts darüber aus, ob    //
//  eine Empfangssequenz noch im Gange ist oder nicht. Dazu siehe BOARD_BusIs-          //
//  Receiving().                                                                        //
//                                                                                      //
//  Eingabe: · byte: Adresse des Bytes, das das nächste zur Abholung bereitstehende     //
//                   Datenbyte aufnimmt.                                                //
//  Ausgabe: 'true', wenn in 'byte' ein Datenbyte abgelegt wurde.                       //
//////////////////////////////////////////////////////////////////////////////////////////
bool
BOARD_BusReceive(UInt08* byte)
{
    bool  hasData = false;
    Atomic
    {   UInt08 head = busRxQueueHead;
        if (!busRxQueue[head].timeout)
        {   // Keine Zeitüberschreitungs-Markierung gesetzt
            if (head != busRxQueueTail)
            {   // Datenbyte vorhanden
                *byte = busRxQueue[head].data;
				head++;
				if (head == CARDINALITY(busRxQueue)) head = 0;
				busRxQueueHead = head;
                hasData = true;
			}
		}
    }
    return hasData;
}

//////////////////////////////////////////////////////////////////////////////////////////
//  BOARD_BusIsReceiving(): Prüft, ob die aktuelle Empfangssequenz noch im Gange ist.   //
//  ----------------------                                                              //
//  Eine Gruppe empfangener Datenbytes gilt als zusammenhängene Empfangssequenz, wenn   //
//  die Zeitspanne zwischen dem Empfang sukzessiver Datenbytes einen bestimmten Zeit-   //
//  rahmen nicht überschreitet. Mit anderen Worten: Liegen die Zeitpunkte zweier auf-   //
//  einanderfolgend empfangener Datenbytes länger als CFG_BUS_INTERBYTE_TIMEOUT Milli-  //
//  sekunden auseinander (timeout), gilt die Empfangssequenz als unterbrochen.          //
//                                                                                      //
//    Empfang: ···-|A|-|B|-|C|---------------×---|D|----|E|---------------×--|F|-···    //
//                           :               :            :               :             //
//                           |--- timeout ---|            |--- timeout ---|             //
//                                                                                      //
//  Die Unterbrechungen werden mit in der Empfangsschlange vermerkt, so dass beim Aus-  //
//  lesen der empfangenen Bytes mittels BOARD_BusReceive() die Empfangssequenzen zeit-  //
//  unabhängig separiert werden können. Solange BOARD_BusIsReceiving() mit 'true' zu-   //
//  rückkehrt, ist die Empfangssequenz noch im Gange.                                   //
//                                                                                      //
//      Beispiel: Empfang von Byte-Sequenzen:                                           //
//                                                                                      //
//       const UInt08*                                                                  //
//       TakeSequence(UInt08* length)                                                   //
//       {                                                                              //
//           static UInt08 buffer[BUFFER_SIZE];     // BUFFER_SIZE: 1 .. 254            //
//           static UInt08 index = 0;                                                   //
//                                                                                      //
//           if (BOARD_BusIsReceiving())                                                //
//           {   // Sequenz läuft: Datenbytes abholen und aufsammeln                    //
//               UInt08 byte;                                                           //
//               while (BOARD_BusReceive(&byte))                                        //
//               {   if (index < BUFFER_SIZE)                                           //
//                   {   buffer[index] = byte;                                          //
//                       index++;                                                       //
//                   }                                                                  //
//                   else                                                               //
//                   {   // Überlauf ...                                                //
//                   }                                                                  //
//               }                                                                      //
//           }                                                                          //
//           else                                                                       //
//           {   // Sequenz beendet                                                     //
//               *length = index;   // Anzahl Datenbytes: 0 .. BUFFER_SIZE              //
//               index = 0;                                                             //
//               if (*length > 0) return buffer; // Sequenz ausgeben                    //
//           }                                                                          //
//           *length = 0;                                                               //
//           return NULL;                                                               //
//       }                                                                              //
//                                                                                      //
//                                                                                      //
//  Eingabe: -                                                                          //
//  Ausgabe: 'true', solange die Empfangssequenz im Gange ist.                          //
//////////////////////////////////////////////////////////////////////////////////////////
bool
BOARD_BusIsReceiving(void)
{
    bool  timeout = false;
    Atomic
    {   UInt08 head = busRxQueueHead;
        timeout = busRxQueue[head].timeout;
        if (head != busRxQueueTail) busRxQueue[head].timeout = false;
    }
    return !timeout;
}

//////////////////////////////////////////////////////////////////////////////////////////
//  BOARD_SerialOnActivity(): Rückruffunktion der Seriellen Übertragungseinheit. Die    //
//  ------------------------  Funktion wird aufgerufen, wenn mit dem Gerät über RS232   //
//  kommuniziert wird.                                                                  //
//                                                                                      //
//  Eingabe: -                                                                          //
//  Ausgabe: -                                                                          //
//////////////////////////////////////////////////////////////////////////////////////////
void
BOARD_SerialOnActivity(void)
{
	serActLedOn = CFG_SER_ACT_LED_FLASH;
}

//////////////////////////////////////////////////////////////////////////////////////////
void
BOARD_SerialTransmit(UInt08 byte)
{
    bool full = false;
    Atomic
    {   UInt08 tail = serTxQueueHead + serTxQueueLength;
        if (CARDINALITY(serTxQueue) > 127)
        {   if      (tail >= CARDINALITY(serTxQueue)) tail -= CARDINALITY(serTxQueue);
            else if (tail < serTxQueueHead)           tail += 256 - CARDINALITY(serTxQueue);
        }
        else
        {   if (tail >= CARDINALITY(serTxQueue)) tail -= CARDINALITY(serTxQueue);
        }
        serTxQueue[tail] = byte;
        serTxQueueLength++;
        full = serTxQueueLength == CARDINALITY(serTxQueue);
        UCSR1B |= BV(UDRIE1);   // "Data Register Empty" Interrupt ein
    }
    while (full)
    {   Atomic { full = serTxQueueLength == CARDINALITY(serTxQueue); }
    }
}

//////////////////////////////////////////////////////////////////////////////////////////
bool
BOARD_SerialReceive(UInt08* byte)
{
    bool  newByte = false;
    Atomic
    {   if (serRxQueueLength == 0)
        {   newByte = false;
        }
        else
        {   *byte = serRxQueue[serRxQueueHead];
            serRxQueueHead++;
            if (serRxQueueHead == CARDINALITY(serRxQueue)) serRxQueueHead = 0;
            serRxQueueLength--;
            newByte = true;
        }
    }
    return newByte;
}

//////////////////////////////////////////////////////////////////////////////////////////
void
BOARD_SerialFlush(void)
{
    bool empty;
    do
    {   Atomic { empty = serTxQueueLength == 0; }
    } while (!empty);
}

//////////////////////////////////////////////////////////////////////////////////////////
//  BOARD_Clock(): Liest die Systemuhr aus.                                             //
//  -------------                                                                       //
//                                                                                      //
//  Eingabe: · dateTime: Referenz, die die Zeitinformation aufnimmt.                    //
//  Ausgabe: -                                                                          //
//////////////////////////////////////////////////////////////////////////////////////////
void
BOARD_Clock(DateTime* dateTime)
{
    *dateTime = sysDateTime;
}

//////////////////////////////////////////////////////////////////////////////////////////
//  BOARD_SetClock(): Stellt die Systemuhr.                                             //
//  ----------------                                                                    //
//                                                                                      //
//  Eingabe: · dateTime: Datumzeit, mit der die Systemuhr gestellt wird.                //
//  Ausgabe: -                                                                          //
//////////////////////////////////////////////////////////////////////////////////////////
void
BOARD_SetClock(const DateTime* dateTime)
{
    sysDateTime = *dateTime;
    if (!DATETIME_IsValid(dateTime)) sysDateTime.time.second = 255;
}

//////////////////////////////////////////////////////////////////////////////////////////
//  BOARD_IsMuted(): Prüft, ob die Tonausgabe abgeschaltet ist.                         //
//  ----------------                                                                    //
//                                                                                      //
//  Eingabe:                                                                            //
//  Ausgabe: 'true': Tonausgabe ist stummgeschaltet.                                    //
//////////////////////////////////////////////////////////////////////////////////////////
bool
BOARD_IsMuted(void)
{
    return toneMute;
}

//////////////////////////////////////////////////////////////////////////////////////////
//  BOARD_SetMuted(): Ändert die Stummschaltung der Tonausgabe.                         //
//  ----------------                                                                    //
//                                                                                      //
//  Eingabe: · mute: 'true'/'false': Tonausgabe ein/aus                                 //
//  Ausgabe: -                                                                          //
//////////////////////////////////////////////////////////////////////////////////////////
void
BOARD_SetMuted(bool mute)
{
    toneMute = mute;
    if (mute) BOARD_PlayTone(NOTE_XX, 0);
}

//////////////////////////////////////////////////////////////////////////////////////////
//  BOARD_PlayTone(): Spielt den angegebenen Ton.                                       //
//  ----------------                                                                    //
//                                                                                      //
//  Eingabe: · note:    Die abzuspielende Tonhöhe als Aufzählungstyp 'Note'.            //
//           · sustain: Haltedauer: 0 .. 250 [1÷100s]                                   //
//  Ausgabe: -                                                                          //
//////////////////////////////////////////////////////////////////////////////////////////
void
BOARD_PlayTone(Note note, UInt08 sustain)
{
    static const UInt08  lookupTable[][2] PROGMEM =   // Tonfrequenz-Registerwerte LUT
    {   // Ni , OCR         Ton  = Frequenz    |   µC-Freq     (Abweichung)    Enum
        //-------------------------------------+----------------------------------------------
        {  0  , 0x00 },  // --                 |                               NOTE_XX  =  0
        {  4  , 0xA4 },  // C3     130.813 Hz  |   130.909 Hz    (+0.07 %)     NOTE_C3  =  1
        {  4  , 0x9B },  // C3#    138.591 Hz  |   138.462 Hz    (-0.09 %)     NOTE_C3f =  2
        {  4  , 0x92 },  // D3     146.832 Hz  |   146.939 Hz    (+0.07 %)     NOTE_D3  =  3
        {  4  , 0x8A },  // D3#    155.563 Hz  |   155.396 Hz    (-0.11 %)     NOTE_D3f =  4
        {  4  , 0x82 },  // E3     164.814 Hz  |   164.885 Hz    (+0.04 %)     NOTE_E3  =  5
        {  4  , 0x7B },  // F3     174.614 Hz  |   174.194 Hz    (-0.24 %)     NOTE_F3  =  6
        {  4  , 0x74 },  // F3#    184.997 Hz  |   184.615 Hz    (-0.21 %)     NOTE_F3f =  7
        {  4  , 0x6D },  // G3     195.998 Hz  |   196.364 Hz    (+0.19 %)     NOTE_G3  =  8
        {  4  , 0x67 },  // G3#    207.652 Hz  |   207.692 Hz    (+0.02 %)     NOTE_G3f =  9
        {  4  , 0x61 },  // A3     220.000 Hz  |   220.408 Hz    (+0.19 %)     NOTE_A3  = 10
        {  4  , 0x5C },  // B3     233.082 Hz  |   232.258 Hz    (-0.35 %)     NOTE_B3  = 11
        {  4  , 0x56 },  // H3     246.942 Hz  |   248.276 Hz    (+0.54 %)     NOTE_H3  = 12
        {  4  , 0x52 },  // C4     261.626 Hz  |   260.241 Hz    (-0.53 %)     NOTE_C4  = 13
        {  4  , 0x4D },  // C4#    277.183 Hz  |   276.923 Hz    (-0.09 %)     NOTE_C4f = 14
        {  4  , 0x49 },  // D4     293.665 Hz  |   291.892 Hz    (-0.60 %)     NOTE_D4  = 15
        {  4  , 0x44 },  // D4#    311.127 Hz  |   313.043 Hz    (+0.62 %)     NOTE_D4f = 16
        {  4  , 0x41 },  // E4     329.628 Hz  |   327.273 Hz    (-0.71 %)     NOTE_E4  = 17
        {  3  , 0xF6 },  // F4     349.228 Hz  |   349.798 Hz    (+0.16 %)     NOTE_F4  = 18
        {  3  , 0xE9 },  // F4#    369.994 Hz  |   369.231 Hz    (-0.21 %)     NOTE_F4f = 19
        {  3  , 0xDB },  // G4     391.995 Hz  |   392.727 Hz    (+0.19 %)     NOTE_G4  = 20
        {  3  , 0xCF },  // G4#    415.305 Hz  |   415.385 Hz    (+0.02 %)     NOTE_G4f = 21
        {  3  , 0xC3 },  // A4     440.000 Hz  |   440.816 Hz    (+0.19 %)     NOTE_A4  = 22
        {  3  , 0xB8 },  // B4     466.164 Hz  |   467.027 Hz    (+0.19 %)     NOTE_B4  = 23
        {  3  , 0xAE },  // H4     493.883 Hz  |   493.714 Hz    (-0.03 %)     NOTE_H4  = 24
        {  3  , 0xA4 },  // C5     523.251 Hz  |   523.636 Hz    (+0.07 %)     NOTE_C5  = 25
        {  3  , 0x9B },  // C5#    554.365 Hz  |   553.846 Hz    (-0.09 %)     NOTE_C5f = 26
        {  3  , 0x92 },  // D5     587.330 Hz  |   587.755 Hz    (+0.07 %)     NOTE_D5  = 27
        {  3  , 0x8A },  // D5#    622.254 Hz  |   621.583 Hz    (-0.11 %)     NOTE_D5f = 28
        {  3  , 0x82 },  // E5     659.255 Hz  |   659.542 Hz    (+0.04 %)     NOTE_E5  = 29
        {  3  , 0x7B },  // F5     698.456 Hz  |   696.774 Hz    (-0.24 %)     NOTE_F5  = 30
        {  3  , 0x74 },  // F5#    739.989 Hz  |   738.462 Hz    (-0.21 %)     NOTE_F5f = 31
        {  3  , 0x6D },  // G5     783.991 Hz  |   785.455 Hz    (+0.19 %)     NOTE_G5  = 32
        {  3  , 0x67 },  // G5#    830.609 Hz  |   830.769 Hz    (+0.02 %)     NOTE_G5f = 33
        {  3  , 0x61 },  // A5     880.000 Hz  |   881.633 Hz    (+0.19 %)     NOTE_A5  = 34
        {  3  , 0x5C },  // B5     932.328 Hz  |   929.032 Hz    (-0.35 %)     NOTE_B5  = 35
        {  3  , 0x56 },  // H5     987.767 Hz  |   993.103 Hz    (+0.54 %)     NOTE_H5  = 36
        {  3  , 0x52 },  // C6    1046.500 Hz  |  1040.964 Hz    (-0.53 %)     NOTE_C6  = 37
        {  3  , 0x4D },  // C6#   1108.730 Hz  |  1107.692 Hz    (-0.09 %)     NOTE_C6f = 38
        {  3  , 0x49 },  // D6    1174.660 Hz  |  1167.568 Hz    (-0.60 %)     NOTE_D6  = 39
        {  3  , 0x44 },  // D6#   1244.510 Hz  |  1252.174 Hz    (+0.62 %)     NOTE_D6f = 40
        {  3  , 0x41 },  // E6    1318.510 Hz  |  1309.091 Hz    (-0.71 %)     NOTE_E6  = 41
        {  3  , 0x3D },  // F6    1396.910 Hz  |  1393.548 Hz    (-0.24 %)     NOTE_F6  = 42
        {  3  , 0x39 },  // F6#   1479.980 Hz  |  1489.655 Hz    (+0.65 %)     NOTE_F6f = 43
        {  3  , 0x36 },  // G6    1567.980 Hz  |  1570.909 Hz    (+0.19 %)     NOTE_G6  = 44
        {  3  , 0x33 },  // G6#   1661.220 Hz  |  1661.538 Hz    (+0.02 %)     NOTE_G6f = 45
        {  3  , 0x30 },  // A6    1760.000 Hz  |  1763.265 Hz    (+0.19 %)     NOTE_A6  = 46
        {  3  , 0x2D },  // B6    1864.660 Hz  |  1878.261 Hz    (+0.73 %)     NOTE_B6  = 47
        {  3  , 0x2B },  // H6    1975.530 Hz  |  1963.636 Hz    (-0.60 %)     NOTE_H6  = 48
        {  3  , 0x28 },  // C7    2093.000 Hz  |  2107.317 Hz    (+0.68 %)     NOTE_C7  = 49
        {  3  , 0x26 },  // C7#   2217.460 Hz  |  2215.385 Hz    (-0.09 %)     NOTE_C7f = 50
        {  3  , 0x24 },  // D7    2349.320 Hz  |  2335.135 Hz    (-0.60 %)     NOTE_D7  = 51
        {  3  , 0x22 },  // D7#   2489.020 Hz  |  2468.571 Hz    (-0.82 %)     NOTE_D7f = 52
        {  3  , 0x20 },  // E7    2637.020 Hz  |  2618.182 Hz    (-0.71 %)     NOTE_E7  = 53
        {  2  , 0xF6 },  // F7    2793.830 Hz  |  2798.381 Hz    (+0.16 %)     NOTE_F7  = 54
        {  2  , 0xE9 },  // F7#   2959.960 Hz  |  2953.846 Hz    (-0.21 %)     NOTE_F7f = 55
        {  2  , 0xDB },  // G7    3135.960 Hz  |  3141.818 Hz    (+0.19 %)     NOTE_G7  = 56
        {  2  , 0xCF },  // G7#   3322.440 Hz  |  3323.077 Hz    (+0.02 %)     NOTE_G7f = 57
        {  2  , 0xC3 },  // A7    3520.000 Hz  |  3526.531 Hz    (+0.19 %)     NOTE_A7  = 58
        {  2  , 0xB8 },  // B7    3729.310 Hz  |  3736.216 Hz    (+0.19 %)     NOTE_B7  = 59
        {  2  , 0xAE },  // H7    3951.070 Hz  |  3949.714 Hz    (-0.03 %)     NOTE_H7  = 60
        {  2  , 0xA4 },  // C8    4186.010 Hz  |  4189.091 Hz    (+0.07 %)     NOTE_C8  = 61
        {  2  , 0x9B },  // C8#   4434.922 Hz  |  4430.769 Hz    (-0.09 %)     NOTE_C8f = 62
        {  2  , 0x92 },  // D8    4698.636 Hz  |  4702.041 Hz    (+0.07 %)     NOTE_D8  = 63
        {  2  , 0x8A },  // D8#   4978.032 Hz  |  4972.662 Hz    (-0.11 %)     NOTE_D8f = 64
        {  2  , 0x82 },  // E8    5274.041 Hz  |  5276.336 Hz    (+0.04 %)     NOTE_E8  = 65
        {  2  , 0x7B },  // F8    5587.652 Hz  |  5574.194 Hz    (-0.24 %)     NOTE_F8  = 66
        {  2  , 0x74 },  // F8#   5919.911 Hz  |  5907.692 Hz    (-0.21 %)     NOTE_F8f = 67
        {  2  , 0x6D },  // G8    6271.927 Hz  |  6283.636 Hz    (+0.19 %)     NOTE_G8  = 68
        {  2  , 0x67 },  // G8#   6644.875 Hz  |  6646.154 Hz    (+0.02 %)     NOTE_G8f = 69
        {  2  , 0x61 },  // A8    7040.000 Hz  |  7053.061 Hz    (+0.19 %)     NOTE_A8  = 70
        {  2  , 0x5C },  // B8    7458.620 Hz  |  7432.258 Hz    (-0.35 %)     NOTE_B8  = 71
        {  2  , 0x56 },  // H8    7902.133 Hz  |  7944.828 Hz    (+0.54 %)     NOTE_H8  = 72
        {  2  , 0x52 }   // C9    8372.018 Hz  |  8327.711 Hz    (-0.53 %)     NOTE_C9  = 73
    };

    if (toneMute) note = NOTE_XX;
	toneSustain = sustain;
	TCCR0A = 0;
    OCR0A  = pgm_read_byte(&lookupTable[(UInt08)note][1]);
	TCNT0  = 0;
    UInt08 i = pgm_read_byte(&lookupTable[(UInt08)note][0]);
    static const UInt08 n[] PROGMEM =
    {   0,                   // [0] Ton aus
        BV(CS00),            // [1]    1
        BV(CS01),            // [2]    8
        BV(CS01) | BV(CS00), // [3]   64
        BV(CS02)             // [4]  256
    };
	UInt08 cs = pgm_read_byte(&n[i]);
	if (cs != 0)
	{	TCCR0A = BV(COM0A0) | BV(WGM01);
		TCCR0B = pgm_read_byte(&n[i]);
	}
}

//////////////////////////////////////////////////////////////////////////////////////////
//  BOARD_IsSilence(): Prüft, ob eine mittels BOARD_PlayTone() initiierte Tonausgabe    //
//  ----------------   noch im Gange ist.                                               //
//                                                                                      //
//  Eingabe:                                                                            //
//  Ausgabe: 'true': Es findet aktuell keine Tonausgabe statt.                          //
//////////////////////////////////////////////////////////////////////////////////////////
bool
BOARD_IsSilence(void)
{
	return toneSustain == 0;
}

//////////////////////////////////////////////////////////////////////////////////////////
//  BOARD_Led1Pwm(): Liefert den aktuelle eingestellten PWM-Wert der LED 1.             //
//  ---------------                                                                     //
//                                                                                      //
//  Eingabe: -                                                                          //
//  Ausgabe: PWM-Wert: 0 .. 255 [#].                                                    //
//////////////////////////////////////////////////////////////////////////////////////////
UInt08
BOARD_Led1Pwm(void)
{
    return OCR2A;
}

//////////////////////////////////////////////////////////////////////////////////////////
//  BOARD_SetLed1Pwm(): Setzt den PWM-Wert von LED 1.                                   //
//  ------------------                                                                  //
//                                                                                      //
//  Eingabe: · pwm: PWM-Wert der LED: 0 .. 255 [#].                                     //
//  Ausgabe: -                                                                          //
//////////////////////////////////////////////////////////////////////////////////////////
void
BOARD_SetLed1Pwm(UInt08 pwm)
{
    OCR2A = pwm;
}

//////////////////////////////////////////////////////////////////////////////////////////
//  BOARD_GpioConfiguration(): Liest die Konfiguriert des angegebenen GPIO-Pins.        //
//  -------------------------                                                           //
//                                                                                      //
//  Eingabe: · gpio: Betroffener GPIO-Pin gem. Aufzählungstyp "Gpio".                   //
//  Ausgabe: Die Konfiguration gem. Aufzählungstyp "PinConf".                           //
//////////////////////////////////////////////////////////////////////////////////////////
PinConf
BOARD_GpioConfiguration(Gpio gpio)
{
    UInt08 v;

    switch (gpio)
    {   case GPIO_A0: v = PortPinConf(A, 0); break;
        case GPIO_A1: v = PortPinConf(A, 1); break;
        case GPIO_A2: v = PortPinConf(A, 2); break;
        case GPIO_A3: v = PortPinConf(A, 3); break;
        case GPIO_B0: v = PortPinConf(B, 0); break;
        case GPIO_B4: v = PortPinConf(B, 4); break;
        case GPIO_D4: v = PortPinConf(D, 4); break;
        default:      v = _P_;
    }
    switch (v)
    {   case _T_: return PIN_T;
        case _P_: return PIN_P;
        case _L_: return PIN_L;
        case _H_: return PIN_H;
    }
    return PIN_P;
}

//////////////////////////////////////////////////////////////////////////////////////////
//  BOARD_ConfigureGpio(): Konfiguriert den angegebenen GPIO-Pin.                       //
//  ---------------------                                                               //
//                                                                                      //
//  Eingabe: · gpio: Betroffener GPIO-Pin gem. Aufzählungstyp "Gpio".                   //
//           · conf: Konfiguration gem. Aufzählungstyp "PinConf".                       //
//  Ausgabe: -                                                                          //
//////////////////////////////////////////////////////////////////////////////////////////
void
BOARD_ConfigureGpio(Gpio gpio, PinConf conf)
{
    switch (gpio)
    {   case GPIO_A0: ConfPortPin(A, 0, conf); break;
        case GPIO_A1:
            switch (conf)
            {   case PIN_T: ConfPortPin(A, 1, _T_); break;
                case PIN_P: ConfPortPin(A, 1, _P_); break;
                case PIN_H: ConfPortPin(A, 1, _H_); break;
                case PIN_L: ConfPortPin(A, 1, _L_); break;
                default:    ;
            }
            break;
        case GPIO_A2:
            switch (conf)
            {   case PIN_T: ConfPortPin(A, 2, _T_); break;
                case PIN_P: ConfPortPin(A, 2, _P_); break;
                case PIN_H: ConfPortPin(A, 2, _H_); break;
                case PIN_L: ConfPortPin(A, 2, _L_); break;
                default:    ;
            }
            break;
        case GPIO_A3:
            switch (conf)
            {   case PIN_T: ConfPortPin(A, 3, _T_); break;
                case PIN_P: ConfPortPin(A, 3, _P_); break;
                case PIN_H: ConfPortPin(A, 3, _H_); break;
                case PIN_L: ConfPortPin(A, 3, _L_); break;
                default:    ;
            }
            break;
        case GPIO_B0:
            switch (conf)
            {   case PIN_T: ConfPortPin(B, 0, _T_); break;
                case PIN_P: ConfPortPin(B, 0, _P_); break;
                case PIN_H: ConfPortPin(B, 0, _H_); break;
                case PIN_L: ConfPortPin(B, 0, _L_); break;
                default:    ;
            }
            break;
        case GPIO_B4:
            switch (conf)
            {   case PIN_T: ConfPortPin(B, 4, _T_); break;
                case PIN_P: ConfPortPin(B, 4, _P_); break;
                case PIN_H: ConfPortPin(B, 4, _H_); break;
                case PIN_L: ConfPortPin(B, 4, _L_); break;
                default:    ;
            }
            break;
        case GPIO_D4:
            switch (conf)
            {   case PIN_T: ConfPortPin(D, 4, _T_); break;
                case PIN_P: ConfPortPin(D, 4, _P_); break;
                case PIN_H: ConfPortPin(D, 4, _H_); break;
                case PIN_L: ConfPortPin(D, 4, _L_); break;
                default:    ;
            }
            break;
        default:
            ;
    }
}

//////////////////////////////////////////////////////////////////////////////////////////
//  BOARD_GPIO(): Liefert den Zustand des angegebenen GPIOs.                            //
//  ------------                                                                        //
//                                                                                      //
//  Eingabe: · gpio: Betroffener GPIO gem. Aufzählungstyp "Gpio".                       //
//  Ausgabe: 'true': Kontakt ist 'high'.                                                //
//////////////////////////////////////////////////////////////////////////////////////////
bool
BOARD_Gpio(Gpio gpio)
{
    switch (gpio)
    {   case GPIO_A0: return !!GetPortPin(A, 0);
        case GPIO_A1: return !!GetPortPin(A, 1);
        case GPIO_A2: return !!GetPortPin(A, 2);
        case GPIO_A3: return !!GetPortPin(A, 3);
        case GPIO_B0: return !!GetPortPin(B, 0);
        case GPIO_B4: return !!GetPortPin(B, 4);
        case GPIO_D4: return !!GetPortPin(D, 4);
        default:      return false;
    }
}

//////////////////////////////////////////////////////////////////////////////////////////
//  BOARD_SetGPIO(): Setzt den Zustand von 'gpio'. Die Funktion kann den Zustand des    //
//  ---------------  nur setzen, wenn der entsprechende Port-Pin als Ausgang konfi-     //
//  guriert ist.                                                                        //
//                                                                                      //
//  Eingabe: · gpio: Betroffener GPIO gem. Aufzählungstyp "Gpio".                       //
//           · on:  'true'/'false' setzt den Pin auf 'high'/'low'.                      //
//  Ausgabe: -                                                                          //
//////////////////////////////////////////////////////////////////////////////////////////
void
BOARD_SetGpio(Gpio gpio, bool on)
{
    switch (gpio)
    {   case GPIO_A0: ChgPortPin(A, 0, on); break;
        case GPIO_A1: ChgPortPin(A, 1, on); break;
        case GPIO_A2: ChgPortPin(A, 2, on); break;
        case GPIO_A3: ChgPortPin(A, 3, on); break;
        case GPIO_B0: ChgPortPin(B, 0, on); break;
        case GPIO_B4: ChgPortPin(B, 4, on); break;
        case GPIO_D4: ChgPortPin(D, 4, on); break;
        default:      ;
    }
}

//////////////////////////////////////////////////////////////////////////////////////////
//  BOARD_AnalogInput(): Liest den Wert des analogen Eingangs aus.                      //
//  -------------------                                                                 //
//                                                                                      //
//  Eingabe: -                                                                          //
//  Ausgabe: Wert des ADCs: 0..1023                                                     //
//////////////////////////////////////////////////////////////////////////////////////////
UInt16
BOARD_AnalogInput(void)
{
    return ReadADC(6);
}

//////////////////////////////////////////////////////////////////////////////////////////
//  BOARD_ReadI2C(): Liest 'count' Datenbytes von einem über I²C angeschlossenen        //
//  ---------------  Gerätes.                                                           //
//                                                                                      //
//  Eingabe: · address: Adresse des I²C-Gerätes: 0 .. 127                               //
//           · buffer:  Zeiger auf einen ausreichend dimensionierten Speicherbereich,   //
//                      der die zu lesenden Bytes aufnimmt.                             //
//           · count:   Anzahl zu lesender Bytes.                                       //
//  Ausgabe: 'true': Lesevorgang war erfolgreich.                                       //
//////////////////////////////////////////////////////////////////////////////////////////
bool
BOARD_ReadI2C(SInt08 address, UInt08* buffer, UInt08 count)
{
    bool result = false;
    // I²C Einheit einschalten und Start-Kennung senden
    TWCR = BV(TWINT) | BV(TWSTA) | BV(TWEN);
    UInt08 t;
    for (t=100; !(TWCR & BV(TWINT)) && t>0; t--) Freeze10us(); // max. 1ms warten
    if (t > 0)
    {   if (TW_STATUS == TW_START)
        {   // Zielgeräte-Addresse mit Leseanforderung übertragen (SLA+R)
            TWDR = ((UInt08)address << 1) | TW_READ;
            TWCR = BV(TWINT) | BV(TWEN);
            for (t=100; !(TWCR & BV(TWINT)) && t>0; t--) Freeze10us();
            if (t > 0)
            {   if (TW_STATUS == TW_MR_SLA_ACK)
                {   // <num> Datenbytes entgegennehmen, das letzte unbestätigt (NACK)
                    UInt08 i;
                    for (i=0; i<count; i++)
                    {   UInt08  acknowledge;
                        if (i < count-1)
                        {   TWCR = BV(TWINT) | BV(TWEN) | BV(TWEA);
                            acknowledge = TW_MR_DATA_ACK;
                        }
                        else
                        {   TWCR = BV(TWINT) | BV(TWEN);
                            acknowledge = TW_MR_DATA_NACK;
                        }
                        for (t=100; !(TWCR & BV(TWINT)) && t>0; t--) Freeze10us();
                        if (t == 0 || TW_STATUS != acknowledge) break;
                        buffer[i] = TWDR;
                    }
                    result = i == count;
                }
            }
        }
    }
    // Stop-Kennung senden und I²C Einheit abschalten
    TWCR = BV(TWINT)|BV(TWSTO)|BV(TWEN);
    Freeze100us();
    TWCR = 0;
    return result;
}

//////////////////////////////////////////////////////////////////////////////////////////
//  BOARD_WriteI2C(): Sendet 'count' Datenbytes aus 'buffer' an ein über I²C ange-      //
//  ----------------  schlossenen Gerätes.                                              //
//                                                                                      //
//  Eingabe: · address: Adresse des I²C-Gerätes: 0 .. 127.                              //
//           · buffer:  Zeiger auf den Speicherbereich mit den zu schreibenden Bytes.   //
//           · count:   Anzahl Bytes in 'buffer'.                                       //
//  Ausgabe: 'true': Schreibvorgang war erfolgreich.                                    //
//////////////////////////////////////////////////////////////////////////////////////////
bool
BOARD_WriteI2C(SInt08 address, const UInt08* buffer, UInt08 count)
{
    bool result = false;
    // I²C Einheit einschalten und Start-Kennung senden
    TWCR = BV(TWINT) | BV(TWSTA) | BV(TWEN);
    UInt08 t;
    for (t=100; !(TWCR & BV(TWINT)) && t>0; t--) Freeze10us(); // max. 1ms warten
    if (t > 0)
    {   if (TW_STATUS == TW_START)
        {   // Zielgeräte-Addresse mit Schreib-Anforderung senden (SLA+W)
            TWDR = ((UInt08)address << 1) | TW_WRITE;
            TWCR = BV(TWINT)|BV(TWEN);
            for (t=100; !(TWCR & BV(TWINT)) && t>0; t--) Freeze10us();
            if (t > 0)
            {   if (TW_STATUS == TW_MT_SLA_ACK)
                {   // <num> Datenbytes übertragen
                    UInt08 i;
                    for (i=0; i<count; i++)
                    {   TWDR = buffer[i];
                        TWCR = BV(TWINT)|BV(TWEN);
                        for (t=100; !(TWCR & BV(TWINT)) && t>0; t--) Freeze10us();
                        if (t == 0 || TW_STATUS != TW_MT_DATA_ACK) break;
                    }
                    result = i == count;
                }
            }
        }
    }
    // Stop-Kennung senden und I²C Einheit abschalten
    TWCR = BV(TWINT)|BV(TWSTO)|BV(TWEN);
    Freeze100us();
    TWCR = 0;
    return result;
}

//////////////////////////////////////////////////////////////////////////////////////////
//  BOARD_Led(): Liefert die Zustand von 'led'.                                         //
//  -----------                                                                         //
//                                                                                      //
//  Eingabe: · led: Betroffene LED gem. Aufzählungstyp "Led".                           //
//  Ausgabe: 'true': LED ist eingeschaltet.                                             //
//////////////////////////////////////////////////////////////////////////////////////////
bool
BOARD_Led(Led led)
{
    switch (led)
    {   case LED_2: return !!GetPin(GP_LED2);
        case LED_3: return !!GetPin(GP_LED3);
        default:    return false;
    }
}

//////////////////////////////////////////////////////////////////////////////////////////
//  BOARD_SetLed(): Setzt den Zustand von 'led'.                                        //
//  --------------                                                                      //
//                                                                                      //
//  Eingabe: · led: Betroffene LED gem. Aufzählungstyp "Led".                           //
//           · on:  'true'/'false' schaltet die LED ein/aus.                            //
//  Ausgabe: -                                                                          //
//////////////////////////////////////////////////////////////////////////////////////////
void
BOARD_SetLed(Led led, bool on)
{
    switch (led)
    {   case LED_2: ChgPin(GP_LED2, on); break;
        case LED_3: ChgPin(GP_LED3, on); break;
        default:    ;
    }
}

//////////////////////////////////////////////////////////////////////////////////////////
//  BOARD_TakeKey(): Liefert 'true', wenn seit der letzten Abfrage mit BOARD_TakeKey()  //
//  ---------------  die Taste 'key' gedrückt und gehalten wurde.                       //
//                                                                                      //
//  Eingabe: · key: Betroffene Taste gem.Aufzählungstyp "Key".                          //
//  Ausgabe: 'true', wenn ein Tastendruck-Ereignis stattgefunden hat.                   //
//////////////////////////////////////////////////////////////////////////////////////////
bool
BOARD_TakeKey(Key key)
{
    bool ret = keyPressedEvent[(UInt08)key];
    keyPressedEvent[(UInt08)key] = false;
    return ret;
}

//////////////////////////////////////////////////////////////////////////////////////////
//  BOARD_Key(): Liefert den aktuellen Zustand der Taste 'key'.                         //
//  -----------                                                                         //
//                                                                                      //
//  Eingabe: · key: Betroffene Taste gem.Aufzählungstyp "Key".                          //
//  Ausgabe: 'true', wenn die Taste gerade gedrückt ist.                                //
//////////////////////////////////////////////////////////////////////////////////////////
bool
BOARD_Key(Key key)
{
    return keyPressedState[(UInt08)key];
}

//////////////////////////////////////////////////////////////////////////////////////////
SysLedSignal
BOARD_SysLedSignal(void)
{
    return sysLed;
}

//////////////////////////////////////////////////////////////////////////////////////////
void
BOARD_SetSysLedSignal(SysLedSignal systemLed)
{
    sysLed = systemLed;
}

//////////////////////////////////////////////////////////////////////////////////////////
//  Prozessaufrufe  //////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
void
BOARD_Startup(void)
{
    if (doNext != DO_STARTUP) Reset();
    DisableInterrupts();
    // Watchdog --------------------------------------------------------------------------
    wdt_enable(WDTO_2S);
    // Ports'n'Pins ----------------------------------------------------------------------
    ConfPort(A, _P_,    // PA.0: A0_GPIO    : Allgemeiner Anschluss
                _P_,    // PA.1: A1_GPIO    : Allgemeiner Anschluss
                _P_,    // PA.2: A2_GPIO    : Allgemeiner Anschluss
                _P_,    // PA.3: A3_GPIO    : Allgemeiner Anschluss
                _T_,    // PA.4: A4_PDN	    : Allgemeiner Pull-Down-Anschluss
                _T_,    // PA.5: A5_PUP     : Allgemeiner Pull-Up-Anschluss
                _T_,    // PA.6: A6_AIN     : Analogeingang 0 .. 5V
                _L_);   // PA.7: A7_SYS_LED : System-Led: L = aus
    ConfPort(B, _P_,    // PB.0: B0_GPIO    : Allgemeiner Anschluss
                _T_,    // PB.1: B1_TA1     : Allgemeine Drucktaste 1, L = gedrückt
                _T_,    // PB.2: B2_TA2     : Allgemeine Drucktaste 2  L = gedrückt
                _L_,    // PB.3: B3_PIEZO   : Piezosignalgeber, Frequenz über Timer 0 PWM
                _P_,    // PB.4: B4_GPIO    : Allgemeiner Anschluss
                _P_,    // PB.5: B5_MOSI    : SPI/ISP
                _P_,    // PB.6: B6_MISO    : SPI/ISP
                _P_);   // PB.7: B7_SCK     : SPI/ISP
    ConfPort(C, _P_,    // PC.0: C0_SCL     : TWI (I²C)
                _P_,    // PC.1: C1_SDA     : TWI (I²C)
                _P_,    // PC.2: C2_TCK     : JTAG
                _P_,    // PC.3: C3_TMS     : JTAG
                _P_,    // PC.4: C4_TDO     : JTAG
                _P_,    // PC.5: C5_TDI     : JTAG
                _L_,    // PC.6: C6_BUS_ACT : Cy4NET-Aktivität, L = keine
                _L_);   // PC.7: C7_SER_ACT : Serielle-Aktivität, L = keine
    ConfPort(D, _P_,    // PD.0: RXD0       : Bus-RXD
                _P_,    // PD.1: TXD0       : Bus-TXD
                _P_,    // PD.2: RXD1       : RS232-RXD
                _P_,    // PD.3: TXD1       : RS232-TXD
                _P_,    // PD.4: D4_GPIO    : Allgemeiner Anschluss
                _L_,    // PD.5: D5_GP_LED3 : Allgemeine LED 3, H = Ein
                _L_,    // PD.6: D6_GP_LED2 : Allgemeine LED 2, H = Ein
                _L_);   // PD.7: D7_GP_LED1 : Allgemeine LED 1, H = Ein (PWM Timer 2)
    // Pause, bis sich alles stabilisiert hat --------------------------------------------
    Freeze100us();
    // ADCs ------------------------------------------------------------------------------
    ADCSRA  = BV(ADEN) | BV(ADPS2) | BV(ADPS1) | BV(ADPS0); // ADC einschalten (F_CPU÷128)
    DIDR0   = BV(ADC6D);                                    // Entspr. Digitaleingänge aus
    ADCSRA |= BV(ADSC);                                     // 1. Abtastung sofort starten
    // Timer/Counter 0 als CTC an OC0 --- (Frequenzgenerator für den Signalgeber) --------
    TCCR0A = BV(COM0A0);
    // Timer 1: Clear Timer on Compare (CTC), Intervall: 1ms (Wiederholrate: 1kHz) -------
    TCCR1B = BV(WGM12) | BV(CS10);
    OCR1A  = F_CPU/1000-1;
    TIMSK1 |= BV(OCIE1A);
    // Timer/Counter 2 als Phase Correct PWM an OC2 mit (F_CPU÷64)÷255÷2 = 338.8 Hz ------
    TCCR2A = BV(COM2A1) | BV(WGM20);
    TCCR2B = BV(CS22);
    // USART 0 19200 Baud, 8 Data, 1 Stop, No Parity, Async., IRQs: RXC/TXC --------------
    UCSR0B = BV(RXCIE0) | BV(TXCIE0) | BV(RXEN0) | BV(TXEN0);
    UBRR0  = F_CPU/16L/19200-1;
    // USART 1 57600 Baud, 8 Data, 1 Stop, No Parity, Async., IRQs: RXC/UDRE -------------
    UCSR1B = BV(RXCIE1) | BV(RXEN1) | BV(TXEN1);
    UBRR1  = F_CPU/16L/57600-1;
    // SPI Master, F_CPU÷2, MSB first, Sample on rising Edge, Mode 3 ---------------------
    //SPCR = BV(SPE)| BV(MSTR) | BV(CPOL) | BV(CPHA);
    //SPSR = BV(SPI2X);
    // I²C -------------------------------------------------------------------------------
    TWCR = BV(TWEN);        // I²C einschalten, Master,
    TWAR = 0;               // "General Call" abschalten. Keine Slave-Adr.
    TWBR = 80;              // SCL freq = F_CPU ÷ (16 + 2·TWBR · 4^TWPS) = 62.8 kHz
    TWSR = 0;
    // Sonstiges -------------------------------------------------------------------------
    EnableInterrupts();
    doNext = DO_PROCEED;
}

//////////////////////////////////////////////////////////////////////////////////////////
bool
BOARD_DoMainLoop(void)
{
    return doNext == DO_PROCEED;
}

//////////////////////////////////////////////////////////////////////////////////////////
void
BOARD_Proceed(void)
{
    ProceedLedSignaling();
    ProceedKeying();
    ProceedSysDateTiming();
    ProceedTonePlaying();
    wdt_reset();
}

//////////////////////////////////////////////////////////////////////////////////////////
void
BOARD_Shutdown(void)
{
	while (BOARD_BusIsTransmitting());
	BOARD_SerialFlush();
    switch (doNext)
    {   case DO_PROCEED:      doNext = DO_SHUTDOWN; return;
        case DO_ENTER_PGMLDR: EnterPgmLdr();        break;
        case DO_STARTUP:      Reset();              break;
        default:              ;
    }
    // Abschaltanweisungen ...
}

//////////////////////////////////////////////////////////////////////////////////////////
