﻿//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  Datum & Zeit                                                                        //
//  _______________________________                                                     //
//                                                                                      //
//  Copyright © 2007 Reimar Grasbon                                                     //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  This software is subjected to the FreeBSD License. Redistribution and use in source //
//  and binary forms, with or without modification, are permitted provided that the     //
//  following conditions are met:                                                       //
//                                                                                      //
//  1. Redistributions of source code must retain the above copyright notice, this list //
//     of conditions and the following disclaimer.                                      //
//                                                                                      //
//  2. Redistributions in binary form must reproduce the above copyright notice, this   //
//     list of conditions and the following disclaimer in the documentation and/or      //
//     other materials provided with the distribution.                                  //
//                                                                                      //
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY //
//  EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED           //
//  WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  //
//  IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,    //
//  INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT  //
//  NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR  //
//  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,   //
//  WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)  //
//  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE          //
//  POSSIBILITY OF SUCH DAMAGE.                                                         //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////

#include <avr/pgmspace.h>
#include "DateTime.h"

//////////////////////////////////////////////////////////////////////////////////////////
//  Lokale Funktionen  ///////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////////////////////////
//  Shift(): Bewegt den Wert in <value> um <shift> Werte unter Berücksichtigung der     //
//  -------  Überlaufgrenze in <modulo>.                                                //
//                                                                                      //
//  Eingabe: · value : Der zu verschiebende Wert: 0 .. 120.                             //
//           · modulo: Überlaufgrenze: 1 .. 120.                                        //
//           · shift : Verschiebewert: -128 .. +127.                                    //
//  Ausgabe: Anzahl der bei der Verschiebung stattgefundenen Über- bzw. Unterläufe:     //
//           -128 .. +127.                                                              //
//////////////////////////////////////////////////////////////////////////////////////////
static SInt08
Shift(UInt08* value, UInt08 modulo, SInt08 shift)
{
    *value += shift;
    SInt08  changeover = 0;
    while (*value >= modulo)
    {   if (shift < 0)
        {   *value += modulo;
            changeover--;
        }
        else
        {   *value -= modulo;
            changeover++;
        }
    }
    return changeover;
}

//////////////////////////////////////////////////////////////////////////////////////////
//  Zeitfunktionen  //////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////////////////////////
//  TIME_IsValid(): Prüft, ob die Einträge in <time> gültig sind. Eine gültige Zeit     //
//  --------------  erstreckt sich von 0:00:00 bis 23:59:59.                            //
//                                                                                      //
//  Eingabe: · time: Die zu prüfende Zeit.                                              //
//  Ausgabe: 'true', wenn die Zeit gültig ist.											//
//////////////////////////////////////////////////////////////////////////////////////////
bool
TIME_IsValid(const Time* time)
{
    return time->hour < 24 && time->minute < 60 && time->second < 60;
}

//////////////////////////////////////////////////////////////////////////////////////////
//  TIME_ShiftSecond(): Verschiebt die Zeit in <time> um die in <shift> angegebenen     //
//  ------------------  Sekunden.                                                       //
//                                                                                      //
//  Eingabe: · time:  Die zu verschiebende Zeit.                                        //
//           · shift: Anzahl Sekunden, um die <time> verschoben wird: -128 .. +127.     //
//  Ausgabe: -                                                                          //
//////////////////////////////////////////////////////////////////////////////////////////
void
TIME_ShiftSecond(Time* time, SInt08 shift)
{
    TIME_ShiftMinute(time, Shift(&time->second, 60, shift));
}

//////////////////////////////////////////////////////////////////////////////////////////
//  TIME_ShiftMinute(): Verschiebt die Zeit in <time> um die in <shift> angegebenen     //
//  ------------------  Minuten.                                                        //
//                                                                                      //
//  Eingabe: · time:  Die zu verschiebende Zeit.                                        //
//           · shift: Anzahl Minuten, um die <time> verschoben wird: -128 .. +127.      //
//  Ausgabe: -                                                                          //
//////////////////////////////////////////////////////////////////////////////////////////
void
TIME_ShiftMinute(Time* time, SInt08 shift)
{
    TIME_ShiftHour(time, Shift(&time->minute, 60, shift));
}

//////////////////////////////////////////////////////////////////////////////////////////
//  TIME_ShiftHour(): Verschiebt die Zeit in <time> um die in <shift> angegebenen       //
//  ----------------  Stunden.                                                          //
//                                                                                      //
//  Eingabe: · time:  Die zu verschiebende Zeit.                                        //
//           · shift: Anzahl Stunden, um die <time> verschoben wird: -128 .. +127.      //
//  Ausgabe: -                                                                          //
//////////////////////////////////////////////////////////////////////////////////////////
void
TIME_ShiftHour(Time* time, SInt08 shift)
{
    Shift(&time->hour, 24, shift);
}

//////////////////////////////////////////////////////////////////////////////////////////
//  TIME_Compare(): Vergleicht die angegebenen Zeiten. Ist <timeA> kleiner bzw. früher  //
//  --------------  als <timeB> kehrt die Funktion mit -1 zurück. Ist <timeA> größer    //
//  bzw. später als <timeB>, ist der Rückgabewert +1. Sind die Zeiten gleich, wird 0    //
//  zurückgegeben:                                                                      //
//                          · timeA < timeB    -->   -1                                 //
//                          · timeA = timeB    -->    0                                 //
//                          · timeA < timeB    -->   +1                                 //
//                                                                                      //
//  Eingabe: · timeA: Erste zu vergleichende Zeit.                                      //
//           · timeB: Zweite zu vergleichende Zeit.                                     //
//  Ausgabe: -1, 0, +1, wenn <timeA> <, ==, > <timeB> ist.                              //
//////////////////////////////////////////////////////////////////////////////////////////
SInt08
TIME_Compare(const Time* timeA, const Time* timeB)
{
    if (timeA->hour   < timeB->hour)   return -1;
    if (timeA->hour   > timeB->hour)   return +1;
    if (timeA->minute < timeB->minute) return -1;
    if (timeA->minute > timeB->minute) return +1;
    if (timeA->second < timeB->second) return -1;
    if (timeA->second > timeB->second) return +1;
    return 0;
}

//////////////////////////////////////////////////////////////////////////////////////////
//  Datumsfunktionen  ////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////////////////////////
//  DATE_IsValid(): Prüft, ob die Einträge in <date> gültig sind. Ein gültiges Datum    //
//  --------------  erstreckt sich von 1.1.1700 bis zum 31.12.9000.                     //
//                                                                                      //
//  Eingabe: · date: Das zu prüfende Datum.                                             //
//  Ausgabe: 'true', wenn das Datum gültig ist.											//
//////////////////////////////////////////////////////////////////////////////////////////
bool
DATE_IsValid(const Date* date)
{
    if (date->year < 1700 || date->year > 9000) return false;
    if (date->month == 0  || date->month >  12) return false;
    if (date->day == 0) return false;
    if (date->day > DATE_DaysInMonth(date->month, date->year) ) return false;
    return true;
}

//////////////////////////////////////////////////////////////////////////////////////////
//  DATE_ShiftDay(): Verschiebt das Datum in <date> um die in <shift> angegebene Anzahl //
//  ---------------  Tage.                                                              //
//                                                                                      //
//  Eingabe: · date:  Das zu verschiebende Datum.                                       //
//           · shift: Anzahl Tage, um die <date> verschoben wird: -128 .. +127.         //
//  Ausgabe: -                                                                          //
//////////////////////////////////////////////////////////////////////////////////////////
void
DATE_ShiftDay(Date* date, SInt08 shift)
{
    UInt08 daysInMonth = DATE_DaysInMonth(date->month, date->year);
    while (shift > 0)
    {   if (date->day == daysInMonth)
        {   DATE_ShiftMonth(date, 1);
            date->day = 1;
            daysInMonth = DATE_DaysInMonth(date->month, date->year);
        }
        else
        {   date->day++;
        }
        shift--;
    }
    while (shift < 0)
    {   if (date->day == 1)
        {   DATE_ShiftMonth(date, -1);
            date->day = DATE_DaysInMonth(date->month, date->year);
        }
        else
        {   date->day--;
        }
        shift++;
    }
}

//////////////////////////////////////////////////////////////////////////////////////////
//  DATE_ShiftMonth(): Verschiebt das Datum in <date> um die in <shift> angegebene      //
//  -----------------  Anzahl Monate.                                                   //
//                                                                                      //
//  Hinweis: Um stets ein gültiges Datum zu erhalten, können Veränderungen des Monats   //
//           auch Veränderungen des Tages nötig machen. Befindet sich z.B. in <date>    //
//           der 30. Jan. 2004 und der Monat wird inkrementiert, erhält <date> das      //
//           Datum 29. Feb. 2004. Das bedeutet, dass die Tageszahl bei Veränderungen    //
//           des Monates gegebenenfalls auf das letztmögliche Datum des neuen Monates   //
//           zurückdatiert wird.                                                        //
//                                                                                      //
//  Eingabe: · date:  Das zu verschiebende Datum.                                       //
//           · shift: Anzahl Monate, um die <date> verschoben wird: -128 .. +127.       //
//  Ausgabe: -                                                                          //
//////////////////////////////////////////////////////////////////////////////////////////
void
DATE_ShiftMonth(Date* date, SInt08 shift)
{
    date->month--;
    SInt08 shiftYear = Shift(&date->month, 12, shift);
    date->month++;
    DATE_ShiftYear(date, shiftYear);
    UInt08 daysInMonth = DATE_DaysInMonth(date->month, date->year);
    if (date->day > daysInMonth) date->day = daysInMonth;
}

//////////////////////////////////////////////////////////////////////////////////////////
//  DATE_ShiftYear(): Verschiebt das Datum in <date> um die in <shift> angegebene       //
//  ----------------  Anzahl Jahre. Die Verschiebung kann im Zeitraum vom 1700 bis zum  //
//  9000 stattfinden und wird gegebenenfalls an diesen Jahresgrenzen gestoppt.          //
//                                                                                      //
//  Hinweis: Befindet sich in <date> der 29. Feb. eines Schaltjahres und wird das Jahr  //
//           auf ein "nicht-Schaltjahr" versetzt, wird der Tag auf den 28. Feb. zurück- //
//           datiert.                                                                   //
//                                                                                      //
//  Eingabe: · date:  Das zu verschiebende Datum.                                       //
//           · shift: Anzahl Jahre, um die <date> verschoben wird: -128 .. +127.        //
//  Ausgabe: -                                                                          //
//////////////////////////////////////////////////////////////////////////////////////////
void
DATE_ShiftYear(Date* date, SInt08 shift)
{
    date->year += shift;
    if      (date->year > 9000) date->year = 9000;
    else if (date->year < 1700) date->year = 1700;
    if (date->month == 2 && date->day == 29 && !DATE_IsLeapYear(date->year))
    {   date->day = 28;
    }
}

//////////////////////////////////////////////////////////////////////////////////////////
//  DATE_Compare(): Vergleicht die angegebenen Daten. Ist <dateA> kleiner bzw. früher   //
//  --------------  als <dateB> kehrt die Funktion mit -1 zurück. Ist <dateA> größer    //
//  bzw. später als <dateB>, ist der Rückgabewert +1. Sind die Zeiten gleich, wird 0    //
//  zurückgegeben:                                                                      //
//                          · dateA < dateB    -->    -1                                //
//                          · dateA = dateB    -->     0                                //
//                          · dateA < dateB    -->    +1                                //
//                                                                                      //
//  Eingabe: · dateA: Erstes zu vergleichendes Datum.                                   //
//           · dateB: Zweites zu vergleichendes Datum.                                  //
//  Ausgabe: -1, 0, +1, wenn <dateA> <, ==, > <dateB> ist.                              //
//////////////////////////////////////////////////////////////////////////////////////////
SInt08
DATE_Compare(const Date* dateA, const Date* dateB)
{
    if (dateA->year  < dateB->year)  return -1;
    if (dateA->year  > dateB->year)  return +1;
    if (dateA->month < dateB->month) return -1;
    if (dateA->month > dateB->month) return +1;
    if (dateA->day   < dateB->day)   return -1;
    if (dateA->day   > dateB->day)   return +1;
    return 0;
}

//////////////////////////////////////////////////////////////////////////////////////////
//  DATE_IsLeapYear(): Prüft, ob das angegebene Jahr ein Schaltjahr ist.                //
//  -----------------                                                                   //
//                                                                                      //
//  Eingabe: · year: Das zu prüfende Jahr:  1700 .. 9000.                               //
//  Ausgabe: 'true', wenn <year> ein Schaltjahr ist.                                    //
//////////////////////////////////////////////////////////////////////////////////////////
bool
DATE_IsLeapYear(UInt16 year)
{
    if (year%4 == 0)
    {   if (year%100 == 0)
        {   if (year%400 == 0) return true;
            return false;
        }
        return true;
    }
    return false;
}

//////////////////////////////////////////////////////////////////////////////////////////
//  DATE_DaysInMonth(): Ermittelt die Anzahl der Tage im angegebenen Monat.             //
//  ------------------                                                                  //
//                                                                                      //
//  Eingabe: · month: Monat: 1 .. 12.                                                   //
//           · year:  Jahr: 1700 .. 9000.                                               //
//  Ausgabe: Anzahl der Tage in <month> im Jahr <year>: 28 .. 31.                       //
//////////////////////////////////////////////////////////////////////////////////////////
UInt08
DATE_DaysInMonth(UInt08 month, UInt16 year)
{
    if (month == 2)
    {   if (DATE_IsLeapYear(year)) return 29;
        else                       return 28;
    }
    if (month == 4 || month == 6 || month == 9 || month == 11) return 30;
    else                                                       return 31;
}

//////////////////////////////////////////////////////////////////////////////////////////
//  DATE_WeekDay(): Errechnet den Wochentag. Als Ergebnis liefert die Funktion eine     //
//  --------------  Ausprägung des Aufzählungstyp 'DTWeekday'.                          //
//                                                                                      //
//  Eingabe: · date: Datum des Tages.                                                   //
//  Ausgabe: Der Wochentag als 'DTWeekday'.                                             //
//////////////////////////////////////////////////////////////////////////////////////////
DTWeekday
DATE_WeekDay(const Date* date)
{
    UInt16 y = date->year;
    if (date->month < 3) y--;
    static const UInt08  weekDayT[] PROGMEM = { 0, 3, 2, 5, 0, 3, 5, 1, 4, 6, 2, 4 };
    y = (y + y/4 - y/100 + y/400 + pgm_read_byte(&weekDayT[date->month-1]) + date->day)%7;
    if (y == 0) y = 7;
    return (DTWeekday)y;
}

//////////////////////////////////////////////////////////////////////////////////////////
//  DATE_YearDay(): Errechnet die Nummer des angegebenen Tages im Jahr beginnend bei    //
//  --------------  1 ( = 1. Januar).                                                   //
//                                                                                      //
//  Eingabe: · date: Datum des Tages.                                                   //
//  Ausgabe: Nummer des Tages im Jahr: 1 .. 366.                                        //
//////////////////////////////////////////////////////////////////////////////////////////
SInt16
DATE_YearDay(const Date* date)
{
    UInt08 month = date->month-1;
    SInt16 yday  = date->day;
    while (month > 0)
    {   yday += DATE_DaysInMonth(month, date->year);
        month--;
    }
    return yday;
}

//////////////////////////////////////////////////////////////////////////////////////////
//  DATE_DayType(): Ermittelt die "Art des Tages", d.h. ob es sich bei dem angegebenen  //
//  --------------  Datum um einen Wochenendtag oder Werktag handelt oder ob der Tag    //
//  <date> ein kirchlicher oder gesetzlichen Feiertag ist. Auch anderweitig besondere   //
//  Tage wie z.B. Silvester werden ausgewiesen. Als Ausgabe liefert die Funktion eine   //
//  Ausprägung des Aufzählungstyps 'DTDayType' zurück.                                  //
//  Der Algorithmus arbeitet nur bis zum Jahr 2199. Darüber hinaus werden nur noch      //
//  Werktage und Wochenenden unterschieden.                                             //
//                                                                                      //
//  Eingabe: · date: Datum des Tages.                                                   //
//  Ausgabe: Art des Tages gem. Aufzählungstyp 'DTDayType'.                             //
//////////////////////////////////////////////////////////////////////////////////////////
DTDayType
DATE_DayType(const Date* date)
{
    if (date->year < 2200)
    {   SInt16 num = DATE_YearDay(date);
        if (num == INVALID) return DT_NORMAL_WORKDAY;
        UInt08 a = (UInt08)(((UInt16)(date->year)+100-2000) % 19);
        UInt08 b = (UInt08)(((UInt16)(date->year)+100-2000) %  4);
        UInt08 c = (UInt08)(((UInt16)(date->year)+103-2000) %  7);
        UInt08 d;
        UInt08 e;
        if (date->year < 2100)
        {   d = (19*a + 24)%30;
            e = ((b<<1) + (c<<2) + 6*d + 5)%7;
        }
        else
        {   d = (19*a + 24)%30;
            e = ((b<<1) + (c<<2) + 6*d + 6)%7;
        }
        Date  tmpDate;
        if ((22 + d + e) <= 31)
        {   tmpDate.day   = 22 + d + e;
            tmpDate.month = 3;
            tmpDate.year  = date->year;
            d = (UInt08)DATE_YearDay(&tmpDate);
        }
        else
        {   tmpDate.day = d + e - 9;
            if      (tmpDate.day == 26)                      tmpDate.day = 19;
            else if (tmpDate.day == 25 && d == 28 && a > 10) tmpDate.day = 18;
            tmpDate.month = 4;
            tmpDate.year  = date->year;
            d = (UInt08)DATE_YearDay(&tmpDate);
        }
        tmpDate.day   = 24;
        tmpDate.month = 12;
        tmpDate.year  = date->year;
        UInt08 f = (UInt08)(DATE_WeekDay(&tmpDate));
        if (f == 7) f = 22;
        else        f = 22-f;
        if (date->month == 1 && date->day == 1)   return DT_NEW_YEAR;
        if (date->month == 1 && date->day == 6)   return DT_EPIPHANY;
        if (d - 52 == num)                        return DT_THURSDAY_BEFORE_LENT;
        if (d - 48 == num)                        return DT_MONDAY_BEFORE_LENT;
        if (d - 47 == num)                        return DT_TUESDAY_BEFORE_LENT;
        if (d - 46 == num)                        return DT_ASH_WEDNESDAY;
        if (d -  3 == num)                        return DT_MAUNDY_THURSDAY;
        if (d -  2 == num)                        return DT_GOOD_FRIDAY;
        if (d      == num)                        return DT_EASTER_SUNDAY;
        if (d + 1  == num)                        return DT_EASTER_MONDAY;
        if (date->month == 5 && date->day ==  1)  return DT_MAY_DAY;
        if (d + 39 == num)                        return DT_ASCENSION_THURSDAY;
        if (d + 49 == num)                        return DT_WHITSUNDAY;
        if (d + 50 == num)                        return DT_WHITMONDAY;
        if (d + 60 == num)                        return DT_CORPUS_CHRISTI;
        if (date->month ==  3 && date->day ==  8) return DT_WOMEMS_DAY;
        if (date->month ==  8 && date->day == 15) return DT_ASSUMPTION_DAY;
        if (date->month ==  9 && date->day == 20) return DT_CHILDRENS_DAY;
        if (date->month == 10 && date->day ==  3) return DT_OCTOBER_3RD;
        if (date->month == 10 && date->day == 31) return DT_REFORMATION_DAY;
        if (date->month == 11 && date->day ==  1) return DT_ALLHALLOWS;
        if (date->month == 11 && date->day ==  f) return DT_REPENTANCE_DAY;
        if (date->month == 12 && date->day ==  6) return DT_NICHOLAS_DAY;
        if (date->month == 12 && date->day == 24) return DT_CHRISTMAS_EVE;
        if (date->month == 12 && date->day == 25) return DT_CHRISTMAS_DAY;
        if (date->month == 12 && date->day == 26) return DT_BOXING_DAY;
        if (date->month == 12 && date->day == 31) return DT_NEW_YEARS_EVE;
    }
    DTWeekday weekDay = DATE_WeekDay(date);
    if (weekDay == DT_SAT || weekDay == DT_SUN) return DT_NORMAL_WEEKEND;
    return DT_NORMAL_WORKDAY;
}

//////////////////////////////////////////////////////////////////////////////////////////
//  Datumzeit-Funktionen  ////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////////////////////////
//  DATETIME_IsValid(): Prüft, ob die Einträge in <dateTime> gültig sind. Eine gültige  //
//  ------------------  Datumzeit erstreckt sich Seitens des Datums vom 1.1.1700 bis    //
//  zum 31.12.9000 und Seitens der Zeit von 0:00:00 bis 23:59:59.                       //
//                                                                                      //
//  Eingabe: · dateTime: Die zu prüfende Datumzeit.                                     //
//  Ausgabe: 'true', wenn die Datumzeit gültig ist.										//
//////////////////////////////////////////////////////////////////////////////////////////
bool
DATETIME_IsValid(const DateTime* dateTime)
{
    return TIME_IsValid(&dateTime->time) && DATE_IsValid(&dateTime->date);
}

//////////////////////////////////////////////////////////////////////////////////////////
//  DATETIME_ShiftSecond(): Verschiebt die Datumzeit in <dateTime> um die in <shift>    //
//  ----------------------  angegebenen Sekunden.                                       //
//                                                                                      //
//  Eingabe: · dateTime: Die zu verschiebende Datumzeit.                                //
//           · shift:    Anzahl Sekunden, um die verschoben wird: -128 .. +127.         //
//  Ausgabe: -                                                                          //
//////////////////////////////////////////////////////////////////////////////////////////
void
DATETIME_ShiftSecond(DateTime* dateTime, SInt08 shift)
{
    DATETIME_ShiftMinute(dateTime, Shift(&dateTime->time.second, 60, shift));
}

//////////////////////////////////////////////////////////////////////////////////////////
//  DATETIME_ShiftMinute(): Verschiebt die Datumzeit in <dateTime> um die in <shift>    //
//  ----------------------  angegebenen Minuten.                                        //
//                                                                                      //
//  Eingabe: · dateTime: Die zu verschiebende Datumzeit.                                //
//           · shift:    Anzahl Minuten, um die verschoben wird: -128 .. +127.          //
//  Ausgabe: -                                                                          //
//////////////////////////////////////////////////////////////////////////////////////////
void
DATETIME_ShiftMinute(DateTime* dateTime, SInt08 shift)
{
    DATETIME_ShiftHour(dateTime, Shift(&dateTime->time.minute, 60, shift));
}

//////////////////////////////////////////////////////////////////////////////////////////
//  DATETIME_ShiftHour(): Verschiebt die Datumzeit in <dateTime> um die in <shift>      //
//  --------------------  angegebenen Stunden.                                          //
//                                                                                      //
//  Eingabe: · dateTime: Die zu verschiebende Datumzeit.                                //
//           · shift:    Anzahl Stunden, um die verschoben wird: -128 .. +127.          //
//  Ausgabe: -                                                                          //
//////////////////////////////////////////////////////////////////////////////////////////
void
DATETIME_ShiftHour(DateTime* dateTime, SInt08 shift)
{
    DATETIME_ShiftDay(dateTime, Shift(&dateTime->time.hour, 24, shift));
}

//////////////////////////////////////////////////////////////////////////////////////////
//  DATETIME_ShiftDay(): Verschiebt die Datumzeit in <dateTime> um die in <shift>       //
//  -------------------  angegebenen Tage.                                              //
//                                                                                      //
//  Eingabe: · dateTime: Die zu verschiebende Datumzeit.                                //
//           · shift:    Anzahl Tage, um die verschoben wird: -128 .. +127.             //
//  Ausgabe: -                                                                          //
//////////////////////////////////////////////////////////////////////////////////////////
void
DATETIME_ShiftDay(DateTime* dateTime, SInt08 shift)
{
    DATE_ShiftDay(&dateTime->date, shift);
}

//////////////////////////////////////////////////////////////////////////////////////////
//  DATETIME_ShiftMonth(): Verschiebt die Datumzeit in <dateTime> um die in <shift>     //
//  ---------------------  angegebenen Monate.                                          //
//                                                                                      //
//  Hinweis: Um stets ein gültiges Datum zu erhalten, können Veränderungen des Monats   //
//           auch Veränderungen des Tages nötig machen. Befindet sich z.B. in <date-    //
//           Time> der 30. Jan. 2004 und der Monat wird inkrementiert, erhält <date-    //
//           Time> das Datum 29. Feb. 2004. Das bedeutet, dass die Tageszahl bei Ver-   //
//           änderungen des Monates gegebenenfalls auf das letztmögliche Datum des      //
//           neuen Monates zurückdatiert wird.                                          //
//                                                                                      //
//  Eingabe: · dateTime: Die zu verschiebende Datumzeit.                                //
//           · shift:    Anzahl Monate, um die verschoben wird: -128 .. +127.           //
//  Ausgabe: -                                                                          //
//////////////////////////////////////////////////////////////////////////////////////////
void
DATETIME_ShiftMonth(DateTime* dateTime, SInt08 shift)
{
    DATE_ShiftMonth(&dateTime->date, shift);
}

//////////////////////////////////////////////////////////////////////////////////////////
//  DATETIME_ShiftYear(): Verschiebt die Datumzeit in <dateTime> um die in <shift>      //
//  --------------------  angegebenen Jahre. Die Verschiebung kann im Zeitraum vom 1700 //
//  bis zum 9000 stattfinden und wird gegebenenfalls an diesen Grenzen gestoppt.        //
//                                                                                      //
//  Hinweis: Befindet sich in <dateTime> der 29. Feb. eines Schaltjahres und wird das   //
//           Datum auf ein "nicht-Schaltjahr" versetzt, wird der Tag auf den 28. Feb.   //
//           zurückdatiert.                                                             //
//                                                                                      //
//  Eingabe: · dateTime: Die zu verschiebende Datumzeit.                                //
//           · shift:    Anzahl Jahre, um die verschoben wird: -128 .. +127.            //
//  Ausgabe: -                                                                          //
//////////////////////////////////////////////////////////////////////////////////////////
void
DATETIME_ShiftYear(DateTime* dateTime, SInt08 shift)
{
    DATE_ShiftYear(&dateTime->date, shift);
}

//////////////////////////////////////////////////////////////////////////////////////////
//  DATETIME_Compare(): Vergleicht die angegebenen Datumzeiten. Ist <dateTimeA> kleiner //
//  ------------------  bzw. früher als <dateTimeB> kehrt die Funktion mit -1 zurück.   //
//  Ist <dateTimeA> größer bzw. später als <dateTimeB>, ist der Rückgabewert +1. Sind   //
//  beide Datumzeiten gleich, wird 0 zurückgegeben:                                     //
//                                                                                      //
//                           · dateTimeA < dateTimeB    -->    -1                       //
//                           · dateTimeA = dateTimeB    -->     0                       //
//                           · dateTimeA > dateTimeB    -->    +1                       //
//                                                                                      //
//  Eingabe: · dateTimeA: Erste zu vergleichende Datumzeit.                             //
//           · dateTimeB: Zweite zu vergleichende Datumzeit.                            //
//  Ausgabe: -1, 0, +1, wenn <dateTimeA> <, ==, > <dateTimeB> ist.                      //
//////////////////////////////////////////////////////////////////////////////////////////
SInt08
DATETIME_Compare(const DateTime* dateTimeA, const DateTime* dateTimeB)
{
    SInt08 cmp = DATE_Compare(&dateTimeA->date, &dateTimeB->date);
    if (cmp != 0) return cmp;
    return TIME_Compare(&dateTimeA->time, &dateTimeB->time);
}

//////////////////////////////////////////////////////////////////////////////////////////
//  DATETIME_DSTState(): Prüft, ob der vorliegende Zeitpunkt der Sommer- oder der       //
//  -------------------  Normalzeit zuzuordnen ist.                                     //
//  Die mitteleuropäische Sommerzeit (MESZ) erstreckt sich vom letzten Sonntag im März  //
//  bis zum letzten Sonntag im Oktober. Dabei wird im März die Zeit um 2:00 Uhr um eine //
//  Stunde vor- und im Oktober um 3:00 um eine Stunde zurückgestellt. Mit dieser Um-    //
//  stellungsmethode existieren zwei Zeiten im Jahr die weder der Sommer- noch der Nor- //
//  malzeit zuzuordnen sind. Die Stunde im März zwischen 2:00 Uhr und 3:00 Uhr exis-    //
//  tiert eigentlich nicht, wohingegen die Stunde im Oktober doppelt vorkommt - einmal  //
//  als Sommerzeit und anschließend noch einmal als Normalzeit.                         //
//  Fällt der in der vorliegenden DateTime beschriebene Zeitpunkt in die Sommerzeit-    //
//  Periode, kehrt die Funktion mit MEST zurück. Liegt der Zeitpunkt in der Normalzeit, //
//  ist der Rückgabewert MET. Ist der Zeitpunkt, wie oben beschrieben, keiner Zeit-     //
//  periode zuzuornen, lautet das Ergebnis UITS bzw. UITA (Frühlung bzw. Herbst).       //
//                                                                                      //
//        1.Jan.           Ende März                  Ende Okt.           31.Dez.       //
//        0:00            2:00   3:00                2:00   3:00           23:59        //
//         +------···------+------+--------···--------+------+------···------+          //
//         |///////////////|      |                   |//////|///////////////|          //
//         +------···------+------+--------···--------+------+------···------+          //
//         :               :      +--------···--------+------+               :          //
//         :               :      |///////////////////|//////|               :          //
//         :               :      +--------···--------+------+               :          //
//         :               :      :                   :      :               :          //
//         :..... MET .....: UITS :....... MEST ......: UITA :..... MET .....:          //
//                                                                                      //
//  Eingabe: · dateTime: Zeitpunkt.                                                     //
//  Ausgabe: DST_MEST/DST_MET, wenn der Zeitpunkt der Sommer-/Normalzeit zuzuordnen     //
//           ist. Kann der Zeitpunkt nicht zugeordnet werden, ist der Rückgabewert      //
//           DST_UITS bzw. DST_UITA.                                                    //
//////////////////////////////////////////////////////////////////////////////////////////
DSTState
DATETIME_DSTState(const DateTime* dateTime)
{
    if (dateTime->date.month > 3 && dateTime->date.month < 10) return DST_MEST;
    UInt08 weekDay = (UInt08)(DATE_WeekDay(&dateTime->date));
    if (weekDay == 7) weekDay = 0;
    if (dateTime->date.month == 3)
    {   if (dateTime->date.day < 25 || dateTime->date.day - weekDay < 25) return DST_MET;
        if (weekDay == 0)
        {   if (dateTime->time.hour < 2) return DST_MET;
            if (dateTime->time.hour < 3) return DST_UITS;
        }
        return DST_MEST;
    }
    if (dateTime->date.month == 10)
    {   if (dateTime->date.day < 25 || dateTime->date.day - weekDay < 25) return DST_MEST;
        if (weekDay == 0)
        {   if (dateTime->time.hour < 2) return DST_MEST;
            if (dateTime->time.hour < 3) return DST_UITA;
        }
        return DST_MET;
    }
    return DST_MET;
}

//////////////////////////////////////////////////////////////////////////////////////////
//  DATETIME_Localize(): Versetzt die angegebene Datumzeit von der UTC in eine lokale   //
//  -------------------  Zeitzone. Angegeben wird dazu die gewünschte "Ziel"-Zeitzone   //
//  (z.B. Deutschland: +1, da UTC + 1) und die Information darüber, ob bei der          //
//  Umrechnung die Sommerzeit (gem. MESZ) berücksichtigt werden soll.                   //
//                                                                                      //
//  Eingabe: · timezone: Differenz zur UTC in Stunden: -11 .. +12.                      //
//           · dst:      'true'/'false' = Berücksichtigung der Sommerzeit bei der Um-   //
//                       rechnung ja/nein.                                              //
//           · dateTime: Die zu verstellende Datumzeit.                                 //
//  Ausgabe: -                                                                          //
//////////////////////////////////////////////////////////////////////////////////////////
void
DATETIME_Localize(SInt08 timezone, bool dst, DateTime* dateTime)
{
    DATETIME_ShiftHour(dateTime, timezone);
    if (dst)
    {   DSTState dstState = DATETIME_DSTState(dateTime);
        if (dstState == DST_MEST || dstState == DST_UITS)
        {   DATETIME_ShiftHour(dateTime, 1);
        }
    }
}

//////////////////////////////////////////////////////////////////////////////////////////
//  DATETIME_Universalize(): Versetzt die angegebene Datumzeit von der lokalen in die   //
//  -----------------------  UTC-Zeitzone. Angegeben wird dazu die lokale Zeitzone      //
//  (z.B. Deutschland +1, da UTC + 1) und die Information, ob bei der Umrechnung die    //
//  Sommerzeit (gem. MESZ) berücksichtigt werden soll.                                  //
//  Bei Umrechnungen mit Berücksichtigung der Sommerzeit ist folgendes zu beachten:     //
//  Liegt die angegebene Ortszeit (unerlaubterweise) in Phase 'DST_UITS', wird sie der  //
//  Winterzeit zugeordnet. Fällt sie (erlaubterweise) in 'DST_UITA' wird sie als        //
//  Sommerzeit betrachtet. Bzgl. der Phasen siehe DATETIME_DSTState()).                 //
//                                                                                      //
//  Eingabe: · timezone: Differenz zur UTC in Stunden: -11 .. +12.                      //
//           · dst:      'true'/'false' = Berücksichtigung der Sommerzeit bei der Um-   //
//                       rechnung ja/nein.                                              //
//           · dateTime: Die zu verstellende Datumzeit.                                 //
//  Ausgabe: -                                                                          //
//////////////////////////////////////////////////////////////////////////////////////////
void
DATETIME_Universalize(SInt08 timezone, bool dst, DateTime* dateTime)
{
    if (dst)
    {   DSTState dstState = DATETIME_DSTState(dateTime);
        if (dstState == DST_MEST || dstState == DST_UITA) timezone++;
    }
    DATETIME_ShiftHour(dateTime, -timezone);
}

//////////////////////////////////////////////////////////////////////////////////////////
//  DATETIME_FromData(): Liest aus <data> die Datumzeit aus. Die Funktion erwartet ein  //
//  -------------------  (mindestens) 7 Byte großes Feld, in dem Datumzeit wie folgt    //
//  abgelegt ist:                                                                       //
//                              data[0] = Jahr Hi-Byte                                  //
//                              data[1] = Jahr Lo-Byte                                  //
//                              data[2] = Monat                                         //
//                              data[3] = Tag                                           //
//                              data[4] = Stunde                                        //
//                              data[5] = Minute                                        //
//                              data[6] = Sekunde                                       //
//                                                                                      //
//  Eingabe: · data:     Bytefeldes, dass die Datumzeit bereitstellt.                   //
//           · dateTime: Referenz, die die erhaltene Datumzeit aufnimmt.                //
//  Ausgabe: -                                                                          //
//////////////////////////////////////////////////////////////////////////////////////////
void       
DATETIME_FromData(const UInt08* data, DateTime* dateTime)
{        
    dateTime->date.year   = ToWord(data[0], data[1]);
    dateTime->date.month  = data[2];
    dateTime->date.day    = data[3];
    dateTime->time.hour   = data[4];
    dateTime->time.minute = data[5];
    dateTime->time.second = data[6];
}

//////////////////////////////////////////////////////////////////////////////////////////
//  DATETIME_ToData(): Schreibt die in <dateTime> übergebene Datumzeit in <data>. Die   //
//  -----------------  Funktion erwartet ein (mindestens) 7 Byte großes Feld in das die //
//  Datumzeit wie folgt abgelegt wird.                                                  //
//                                                                                      //
//                              data[0] = Jahr Hi-Byte                                  //
//                              data[1] = Jahr Lo-Byte                                  //
//                              data[2] = Monat                                         //
//                              data[3] = Tag                                           //
//                              data[4] = Stunde                                        //
//                              data[5] = Minute                                        //
//                              data[6] = Sekunde                                       //
//                                                                                      //
//  Eingabe: · dateTime: Die zu übertragende Datumzeit.                                 //
//           · data:     Bytefeldes, dass die Datumzeit aufnimmt.                       //
//  Ausgabe: -                                                                          //
//////////////////////////////////////////////////////////////////////////////////////////
void       
DATETIME_ToData(const DateTime* dateTime, UInt08* data)
{    
    data[0] = HiByte(dateTime->date.year);
    data[1] = LoByte(dateTime->date.year);
    data[2] = dateTime->date.month;
    data[3] = dateTime->date.day;
    data[4] = dateTime->time.hour;
    data[5] = dateTime->time.minute;
    data[6] = dateTime->time.second;
}

//////////////////////////////////////////////////////////////////////////////////////////
